"""Enforce fit limit.

This module was adapted from https://github.com/sfalkner/pynisher
"""

import resource
import signal
import multiprocessing
import os
import sys
import time
import tempfile

import psutil


class CpuTimeoutException(Exception):
    pass


class TimeoutException(Exception):
    pass


class MemorylimitException(Exception):
    pass


class SubprocessException(Exception):
    pass


class AnythingException(Exception):
    pass


# create the function the subprocess can execute
def subprocess_func(
        func, pipe, logger, mem_in_mb, cpu_time_limit_in_s,
        wall_time_limit_in_s, num_procs, grace_period_in_s, tmp_dir,
        *args, **kwargs):

    # simple signal handler to catch the signals for time limits
    def handler(signum, frame):
        # logs message with level debug on this logger
        logger.debug("signal handler: %i" % signum)
        if (signum == signal.SIGXCPU):
            # when process reaches soft limit
            # --> a SIGXCPU signal is sent (it normally terminats the process)
            raise(CpuTimeoutException)
        elif (signum == signal.SIGALRM):
            # SIGALRM is sent to process when the specified time limit to an
            # alarm function elapses (when real or clock time elapses)
            logger.debug("timeout")
            raise(TimeoutException)
        raise AnythingException

    # temporary directory to store stdout and stderr
    if tmp_dir is not None:
        logger.debug(
            'Redirecting output of the function to files. Access them via the '
            'stdout and stderr attributes of the wrapped function.')

        stdout = open(os.path.join(tmp_dir, 'std.out'), 'a', buffering=1)
        sys.stdout = stdout

        stderr = open(os.path.join(tmp_dir, 'std.err'), 'a', buffering=1)
        sys.stderr = stderr

    # catching all signals at this point turned out to interfere with the
    # subprocess (e.g. using ROS)
    signal.signal(signal.SIGALRM, handler)
    signal.signal(signal.SIGXCPU, handler)
    signal.signal(signal.SIGQUIT, handler)

    # code to catch EVERY catchable signal (even X11 related ones ... )
    # only use for debugging/testing as this seems to be too intrusive.
    """
    for i in [x for x in dir(signal) if x.startswith("SIG")]:
        try:
            signum = getattr(signal,i)
            print("register {}, {}".format(signum, i))
            signal.signal(signum, handler)
        except:
            print("Skipping %s"%i)
    """

    # set the memory limit
    if mem_in_mb is not None:
        # byte --> megabyte
        mem_in_b = mem_in_mb * 1024 * 1024
        # the maximum area (in bytes) of address space which may be taken by
        # the process.
        resource.setrlimit(resource.RLIMIT_AS, (mem_in_b, mem_in_b))

    # for now: don't allow the function to spawn subprocesses itself.
    # resource.setrlimit(resource.RLIMIT_NPROC, (1, 1))
    # Turns out, this is quite restrictive, so we don't use this option by
    # default
    if num_procs is not None:
        resource.setrlimit(resource.RLIMIT_NPROC, (num_procs, num_procs))

    # schedule an alarm in specified number of seconds
    if wall_time_limit_in_s is not None:
        signal.alarm(wall_time_limit_in_s)

    if cpu_time_limit_in_s is not None:
        # From the Linux man page:
        # When the process reaches the soft limit, it is sent a SIGXCPU signal.
        # The default action for this signal is to terminate the process.
        # However, the signal can be caught, and the handler can return control
        # to the main program. If the process continues to consume CPU time,
        # it will be sent SIGXCPU once per second until the hard limit is
        # reached, at which time it is sent SIGKILL.
        resource.setrlimit(
            resource.RLIMIT_CPU,
            (cpu_time_limit_in_s, cpu_time_limit_in_s + grace_period_in_s))

    # the actual function call
    try:
        logger.debug("call function")
        return_value = ((func(*args, **kwargs), 0))
        logger.debug("function returned properly: {}".format(return_value))
    except MemoryError:
        return_value = (None, MemorylimitException)

    except OSError as e:
        if (e.errno == 11):
            return_value = (None, SubprocessException)
        else:
            return_value = (None, AnythingException)

    except CpuTimeoutException:
        return_value = (None, CpuTimeoutException)

    except TimeoutException:
        return_value = (None, TimeoutException)

    except AnythingException as e:
        return_value = (None, AnythingException)
    except Exception:
        raise
        logger.debug("Some wired exception occured!")

    finally:
        try:
            logger.debug("return value: {}".format(return_value))

            pipe.send(return_value)
            pipe.close()

        except Exception:
            # this part should only fail if the parent process is already dead
            # so there is not much to do anymore :)
            pass
        finally:
            # recursively kill all children
            p = psutil.Process()
            for child in p.children(recursive=True):
                child.kill()


class enforce_limits (object):

    def __init__(
            self, mem_in_mb=None, cpu_time_in_s=None, wall_time_in_s=None,
            num_processes=None, grace_period_in_s=None, logger=None,
            capture_output=False):
        self.mem_in_mb = mem_in_mb
        self.cpu_time_in_s = cpu_time_in_s
        self.num_processes = num_processes
        self.wall_time_in_s = wall_time_in_s
        self.grace_period_in_s = 0 if grace_period_in_s is None \
            else grace_period_in_s
        self.logger = logger if logger is not None \
            else multiprocessing.get_logger()
        self.capture_output = capture_output

        if self.mem_in_mb is not None:
            self.logger.debug(
                "Restricting your function to {} mb memory."
                .format(self.mem_in_mb))
        if self.cpu_time_in_s is not None:
            self.logger.debug(
                "Restricting your function to {} seconds cpu time."
                .format(self.cpu_time_in_s))
        if self.wall_time_in_s is not None:
            self.logger.debug(
                "Restricting your function to {} seconds wall time."
                .format(self.wall_time_in_s))
        if self.num_processes is not None:
            self.logger.debug(
                "Restricting your function to {} threads/processes."
                .format(self.num_processes))
        if self.grace_period_in_s is not None:
            self.logger.debug(
                "Allowing a grace period of {} seconds."
                .format(self.grace_period_in_s))

    def __call__(self, func):

        class function_wrapper(object):
            def __init__(self2, func):
                self2.func = func
                self2._reset_attributes()

            def _reset_attributes(self2):
                self2.result = None
                self2.exit_status = None
                self2.resources_function = None
                self2.resources_parent = None
                self2.wall_clock_time = None
                self2.stdout = None
                self2.stderr = None

            def __call__(self2, *args, **kwargs):

                self2._reset_attributes()

                # create a pipe to retrieve the return value
                parent_conn, child_conn = multiprocessing.Pipe(False)

                if self.capture_output:
                    tmp_dir = tempfile.TemporaryDirectory()
                    tmp_dir_name = tmp_dir.name

                else:
                    tmp_dir_name = None

                # create and start the process
                subproc = multiprocessing.Process(
                    target=subprocess_func,
                    name="subprocess function call",
                    args=(
                        self2.func,
                        child_conn,
                        self.logger,
                        self.mem_in_mb,
                        self.cpu_time_in_s,
                        self.wall_time_in_s,
                        self.num_processes,
                        self.grace_period_in_s,
                        tmp_dir_name) + args,
                    kwargs=kwargs)
                self.logger.debug(
                    "Function called with argument: {}, {}"
                    .format(args, kwargs))

                # start the process
                start = time.time()
                subproc.start()
                child_conn.close()

                try:
                    # read the return value
                    if (self.wall_time_in_s is not None):
                        if parent_conn.poll(
                                self.wall_time_in_s + self.grace_period_in_s):
                            self2.result, self2.exit_status = \
                                parent_conn.recv()
                        else:
                            subproc.terminate()
                            self2.exit_status = TimeoutException

                    else:
                        self2.result, self2.exit_status = parent_conn.recv()

                except EOFError:    # Don't see that in the unit tests :(
                    self.logger.debug(
                        "Your function call closed the pipe prematurely -> "
                        "Subprocess probably got an uncatchable signal.")
                    self2.exit_status = AnythingException

                except Exception:
                    self.logger.debug("Something else went wrong, sorry.")
                finally:
                    self2.resources_function = resource.getrusage(
                        resource.RUSAGE_CHILDREN)
                    self2.resources_parent = resource.getrusage(
                        resource.RUSAGE_SELF)
                    self2.wall_clock_time = time.time()-start
                    self2.exit_status = 5 if self2.exit_status is None \
                        else self2.exit_status

                    # recover stdout and stderr if requested
                    if self.capture_output:
                        with open(
                                os.path.join(tmp_dir.name, 'std.out'), 'r'
                                ) as fh:
                            self2.stdout = fh.read()
                        with open(
                                os.path.join(tmp_dir.name, 'std.err'), 'r'
                                ) as fh:
                            self2.stderr = fh.read()

                        tmp_dir.cleanup()

                    # don't leave zombies behind
                    subproc.join()

                return (self2.result)

        return (function_wrapper(func))


if __name__ == "__main__":

    import logging

    logging.basicConfig(
        level=logging.INFO,
        format="%(asctime)s: %(levelname)s: %(name)s: %(message)s")

    logger = logging.getLogger(__name__)

    def fit_function(time_in_secs):
        print("fake model fit")
        time.sleep(time_in_secs)
        return time_in_secs

    wrapped_fit_function = enforce_limits(
        wall_time_in_s=3, logger=logger)(fit_function)

    for t in [1, 10]:
        print("result:", wrapped_fit_function(t))
        print("expection:", wrapped_fit_function.exit_status)
        print()



"""Define algorithm component."""

import itertools

import numpy as np

from collections import OrderedDict
from sklearn.base import BaseEstimator
from typing import List, Tuple, Any, Union

from ..data_types import AlgorithmType


Hyperparameters = List[List[Tuple[str, Any]]]

# this variable indicates that a hyperparameter should be completely ignored.
EXCLUDE_ALL = "ALL"


def _check_hyperparameters(
        hyperparameters: Hyperparameters,
        exclusion_conditions: dict) -> Union[dict, None]:
    exclude_cases = {}
    for key, value in hyperparameters:

        if key in exclude_cases and value in exclude_cases[key]:
            # early return: the hyperparameter setting is invalid if any of
            # the selected values are specified in the exclude cases
            return None

        exclude_dict = exclusion_conditions.get(key, None)
        if exclude_dict is None:
            continue
        elif value not in exclude_dict:
            continue
        exclude_cases.update(exclude_dict[value])

    return dict(hyperparameters)


class AlgorithmComponent(object):
    """A component of a machine learning framework."""

    def __init__(
            self,
            name,
            component_class,
            component_type=None,
            initialize_component=None,
            hyperparameters=None,
            constant_hyperparameters=None,
            exclusion_conditions=None):
        """Initialize an AlgorithmComponent.

        :param str name: name of component.
        :param object component_class: of type sklearn.BaseEstimator
        :param str component_type: type of algorithm.
        :param callable|None initialize_component: a function invoked in the
            __call__ method that's responsible for instantiating the
            component_class, with the following signature:

            (component_class: EstimatorClass,
             categorical_features: List[int],
             continuous_features: List[int]) -> EstimatorObject

            This function is called in the __call__ method. If None, then
            component_class Estimator is initialized with all of its defaults.

        :param list[Hyperparameters]|None hyperparameters: list of
            Hyperparameter objects, which specify algorithms' hyperparameter
            space.

            The MetaLearnController interprets `hyperparameters` as a DAG where
            at each node represents a hyperparameter in algorithm space.
            At each node, the agent selects a hyperparameter value (currently
            these can only be a discrete set of values). The agent then
            continues to traverse the graph until all hyperparameters have
            been selected.
        :param dict constant_hyperparameters: a set of hyperparameters that
            shouldn't be picked by the controller
        :param dict[str -> dict[str -> list]] exclusion_conditions: a
            dictionary specifying the which of the subsequent hyperparameters
            should be excluded conditioned on picking a particular
            hyperparameter value. For example:

            {
                "hyperparam1": {
                    "h1_value1": {
                        "hyperparam2": ["h2_value1", "h2_value2"]
                    }
                },
                "hyperparam2": {
                    "h2_value_3": {
                        "hyperparam3": ["h3_value1"]
                    }
                }
            }

            The exclusion criteria are executed in the order specified by
            `hyperparameters`. In the example, if the agent selects
            "h1_value1", then, when it comes to the agent selecting a value
            for `hyperparam2`, the agent will only be able to sample from
            the action probabilities that are not `"h2_value1"` or
            `"h2_value2"`.

            These "action probability masks" prevent the agent from selecting
            actions that the environment (in this case the AlgorithmSpace)
            knows is an invalid combination of hyperparameters.

        """
        if component_type not in AlgorithmType:
            raise ValueError("%s is not a valid algorithm type: choose %s" % (
                component_type, list(AlgorithmType)))
        self.name = name
        self.component_class = component_class
        self.initialize_component = initialize_component
        self.component_type = component_type
        self.hyperparameters = hyperparameters
        self.constant_hyperparameters = {} if \
            constant_hyperparameters is None else constant_hyperparameters
        # TODO: consider making a class for exclusion conditions
        self.exclusion_conditions = exclusion_conditions
        # TODO: implement checker for initialize_component function. Make
        #       sure that the estimator portion of the `transformer` arg
        #       is an Estimator

    def __call__(self, categorical_features=None, continuous_features=None):
        """Instantiate the algorithm object.

        :param dict kwargs: keyword arguments to pass into initialize_component
        """
        if self.initialize_component is None:
            return self.component_class()

        estimator = self.initialize_component(
            self.component_class,
            [] if categorical_features is None else categorical_features,
            [] if continuous_features is None else continuous_features)
        if not isinstance(estimator, BaseEstimator):
            raise TypeError(
                "expected output of `initialize_component` to be a "
                f"BaseEstimator, found {type(estimator)}")
        return estimator

    def get_constant_hyperparameters(self):
        """Get environment-dependent hyperparameters."""
        return {
            "%s__%s" % (self.name, h): value
            for h, value in self.constant_hyperparameters.items()
        }

    def hyperparameter_name_space(self):
        """Return list of hyperparameter names."""
        if self.hyperparameters is None:
            return None
        return ["%s__%s" % (self.name, h.hname) for h in self.hyperparameters]

    def hyperparameter_state_space(self, with_none_token=False):
        """Return dict of hyperparameter space."""
        if self.hyperparameters is None:
            return OrderedDict()
        return OrderedDict([
            (n, h.get_state_space(with_none_token)) for n, h in
            zip(self.hyperparameter_name_space(), self.hyperparameters)])

    def hyperparameter_iterator(self):
        """Return a generator of all possible hyperparameter combinations."""

        expanded_state_space = []
        for key, values in self.hyperparameter_state_space().items():
            expanded_state_space.append([(key, v) for v in values])
        return filter(None, (
            _check_hyperparameters(
                hsetting, self.hyperparameter_exclusion_conditions())
            for hsetting in itertools.product(*expanded_state_space)))

    def hyperparameter_exclusion_conditions(self):
        """Get the conditional map of which hyperparameters go together."""
        if self.hyperparameters is None or self.exclusion_conditions is None:
            return OrderedDict()

        # TODO: make sure that keys are actually hyperparameter names
        def format_exclusion_conditions(conds):
            return {h: {"%s__%s" % (self.name, k): v for k, v in ex.items()}
                    for h, ex in conds.items()}

        return OrderedDict([
            ("%s__%s" % (self.name, hname),
                format_exclusion_conditions(exclusion_conditions))
            for hname, exclusion_conditions
            in self.exclusion_conditions.items()])

    def sample_hyperparameter_state_space(self):
        """Return a random sample from the hyperparameter state space."""
        # TODO: incorporate the exclusion criteria in sampling the state space
        # such that only valid combinations of hyperparameters are given.
        settings = {}
        for key, values in self.hyperparameter_state_space().items():
            settings[key] = values[np.random.randint(len(values))]
        return settings

    def __repr__(self):
        return "<AlgorithmComponent: \"%s\">" % self.name

    def __hash__(self):
        return hash((self.name, self.component_class, self.component_type))

    def __eq__(self, other):
        return hash(self) == hash(other)



"""Track data during hyperparameter search."""

from collections import defaultdict


class TrackerBase(object):

    def __init__(self, metric_names):
        self.metric_names = metric_names
        self.metrics = defaultdict(list)

    def update_metric(self, key: str, value):
        if key not in self.metric_names:
            raise ValueError("metric %s not recognized. Must be one of: {%s}" %
                             (key, self.metric_names))
        self.metrics[key].append(value)

    def update_metrics(self, metrics_dict: dict):
        for key, value in metrics_dict.items():
            self.update_metric(key, value)


class MetricsTracker(TrackerBase):

    def __init__(self):
        super(MetricsTracker, self).__init__([
            "episode",
            "data_env_names",
            "target_type",
            "scorers",
            "total_losses",
            "actor_losses",
            "critic_losses",
            "entropy_losses",
            "gradient_norms",
            "mean_rewards",
            "mean_validation_scores",
            "std_validation_scores",
            "best_validation_scores",
            "best_mlfs",
            "n_successful_mlfs",
            "n_unique_mlfs",
            "n_unique_hyperparams",
            "mlf_diversity",
            "hyperparam_diversity",
            "entropy_coefficient",
        ])

    @property
    def history(self):
        return {k: v for k, v in self.metrics.items() if k != "best_mlfs"}

    @property
    def best_mlfs(self):
        return self.metrics["best_mlfs"]



"""Module that defines the data environment data types."""

from enum import Enum


class AlgorithmType(Enum):

    ONE_HOT_ENCODER = "one_hot_encoder"
    IMPUTER = "imputer"
    RESCALER = "rescaler"
    FEATURE_PREPROCESSOR = "feature_preprocessor"
    CLASSIFIER = "classifier"
    REGRESSOR = "regressor"
    ESTIMATOR = "estimator"


class FeatureType(Enum):
    """Feature type definition, effects the MLF proposal path."""

    CATEGORICAL = 1
    CONTINUOUS = 2
    DATE = 3
    # TODO: Going to hold off on these two types of features since supporting
    # it would add more complexity to the system. Implement this when the
    # classification systems using continuous and categorical data are working.
    # STRING (for text data)


class TargetType(Enum):
    """Target type definition, effects MLF proposal path."""

    BINARY = 1
    MULTICLASS = 2
    REGRESSION = 3
    MULTIREGRESSION = 4


class OpenMLTaskType(Enum):
    """Define task type and ids according to openml docs.

    https://openml.github.io/openml-python/dev/generated/openml.tasks.list_tasks.html#openml.tasks.list_tasks
    """

    # Note: only using SUPERVISED_REGRESSION at the moment
    SUPERVISED_CLASSIFICATION = 1
    SUPERVISED_REGRESSION = 2
    LEARNING_CURVE = 3
    SUPERVISED_DATA_STREAM_CLASSIFICATION = 4
    CLUSTERING = 5
    MACHINE_LEARNING_CHALLENGE = 6
    SURVIVAL_ANALYSIS = 7
    SUBGROUP_DISCOVERY = 8


class DataSourceType(Enum):
    """Define a type of data sources."""

    SKLEARN = 1
    OPEN_ML = 2
    KAGGLE = 3
    AUTOSKLEARN_BENCHMARK = 4
    KIOSK = 5
    OPEN_ML_BENCHMARK_CC18 = 6


class CASHComponent(Enum):
    """A component that the CASH controller predicts to form MLF."""

    ALGORITHM = 1
    HYPERPARAMETER = 2


class ExperimentType(Enum):
    """Experiment types."""

    METALEARN_REINFORCE = 1



"""Utility functions."""

import logging
import time

import math
import numpy as np
import torch

import torch.nn as nn

from .data_types import CASHComponent
from .data_environments.data_environment import NULL_DATA_ENV


def add_metafeatures_hidden_units(controller):

    """Util function that adds hidden units to the metafeature embedding."""
    weights = controller.metafeature_dense.weight.data
    new_w = torch.zeros([weights.shape[0], 1])

    # initialize using default method in pytorch:
    # https://github.com/pytorch/pytorch/blob/master/torch/nn/modules/linear.py#L44-L48  # noqa
    nn.init.kaiming_uniform(new_w, a=math.sqrt(5))
    new_weights = torch.cat([weights, new_w], dim=1)

    controller.metafeature_dense = nn.Linear(
        new_weights.shape[1], controller.metafeature_encoding_size)
    controller.metafeature_dense.weight.data = torch.Tensor(new_weights)
    assert all(controller.metafeature_dense.weight.data[:, -1] ==
               new_w.view(-1))
    return controller


def create_metafeature_spec(data_distribution, scorer_distribution=None):
    """Create a metafeature spec.

    NOTE: may need to make this a class if it becomes more complex.
    """
    if scorer_distribution is not None:
        # does it make sense to have a NULL_SCORER?
        scorer_distribution = (
            "scorer_distribution", str, [s.name for s in scorer_distribution])
    return list(filter(None, [
        ("number_of_examples", int, None),
        ("number_of_features", int, None),
        ("data_env_name", str,
            [d.name for d in data_distribution] + [NULL_DATA_ENV]),
        scorer_distribution,
    ]))


def get_metafeatures_dim(metafeatures_spec):
    """Get dimensionality of metafeatures."""
    return sum([len(m[2]) if m[1] is str else 1 for m in metafeatures_spec])


def init_logging(log_path="/tmp/metalearn.log", log_level=logging.INFO):
    """Initialize logging at the module level.

    :param str|None log_path: path to write logs. If None, write to stdout
    """
    # remove other logging handlers to re-configure logging.
    for h in logging.root.handlers:
        logging.root.removeHandler(h)
    if log_path is not None:
        print("writing logs to %s" % log_path)
        with open(log_path, "w") as f:
            f.close()
    logging.basicConfig(
        filename=log_path,
        level=log_level,
        format="%(asctime)s: %(levelname)s: %(name)s: %(message)s")


class Timer(object):
    """A light-weight timer class."""

    def __enter__(self):
        self.start = time.clock()
        return self

    def __exit__(self, *args):
        self.interval = time.clock() - self.start


def load_model(path, model_class, *args, **kwargs):
    """Load a pytorch model."""
    rnn = model_class(*args, **kwargs)
    rnn.load_state_dict(torch.load(path))
    return rnn


def freeze_model(model):
    """Freeze all model weights to prevent training."""
    for param in model.parameters():
        param.requires_grad = False
    return model


def models_are_equal(x, y):
    return x.config == y.config and \
        len(x.state_dict()) == len(y.state_dict()) and \
        all((
            k1 == k2 and (v1 == v2).all() for (k1, v1), (k2, v2) in
            zip(x.state_dict().items(), y.state_dict().items())))


def aux_tensor(prev_reward):
    """Create an auxiliary input tensor for previous reward and action.

    This is just the reward from the most recent iteration. At the beginning
    of each episode, the previous reward is reset to 0.
    """
    r_tensor = torch.zeros(1, 1, 1)
    return r_tensor + prev_reward


def get_mlf_components(actions):
    algorithms = []
    hyperparameters = {}
    for action in actions:
        if action["action_type"] == CASHComponent.ALGORITHM:
            algorithms.append(action["choice"])
        elif action["action_type"] == CASHComponent.HYPERPARAMETER:
            hyperparameters[action["action_name"]] = action["choice"]
        else:
            raise ValueError(
                "action_type %s not recognized" % action["action_type"])
    return algorithms, hyperparameters


def _create_metafeature_tensor(metafeatures, seq, metafeature_spec):
    """Convert a metafeature vector into a tensor.

    :returns Tensor: dim <string_length x 1 x metafeature_dim>, where
        metafeature_dim is a continuous feature.
    """
    m = []
    for i, feature in enumerate(metafeatures):
        fname, ftype, flevels = metafeature_spec[i]
        if ftype is int:
            metafeature_dim = 1
            feature_val = ftype(feature)
            feature_index = 0
        elif ftype is str:
            metafeature_dim = len(flevels)
            feature_val = 1
            feature_index = flevels.index(feature)
        else:
            raise ValueError(
                "metafeature type %s not recognized" % ftype)
        t = torch.zeros(len(seq), 1, metafeature_dim)
        for j, _ in enumerate(seq):
            t[j][0][feature_index] = feature_val
        m.append(t)
    m = torch.cat(m, 2)
    return m


def _create_input_tensor(a_space, seq):
    """Convert sequence to algorithm space.

    Returns tensor of dim <string_length x 1 x n_components>.

    :param AlgorithmSpace a_space: an algorithm space object that specifies
        the space of possible ML framework components and hyperparameters
    :param list[str] seq: sequence of algorithm components to encode into
        integer tensor.
    :returns: integer algorithm input tensor.
    """
    t = torch.zeros(len(seq), 1, a_space.n_components)
    for i, action in enumerate(seq):
        t[i][0][a_space.components.index(action)] = 1
    return t


def _ml_framework_string(ml_framework):
    return " > ".join(s[0] for s in ml_framework.steps)


def _diversity_metric(n_unique, n_total):
    return np.nan if n_total == 0 else \
        0 if n_total == 1 else \
        (n_unique - 1) / (n_total - 1)


def _hyperparameter_string(hyperparameters):
    return "[%d hyperparams]" % len(hyperparameters)



"""Classifier components."""

from sklearn.ensemble import AdaBoostClassifier, ExtraTreesClassifier, \
    GradientBoostingClassifier, RandomForestClassifier
from sklearn.gaussian_process import GaussianProcessClassifier
from sklearn.linear_model import LogisticRegression
from sklearn.naive_bayes import GaussianNB, MultinomialNB
from sklearn.neighbors import KNeighborsClassifier
from sklearn.svm import LinearSVC, SVC
from sklearn.tree import DecisionTreeClassifier

from .algorithm_component import AlgorithmComponent, EXCLUDE_ALL
from .hyperparameter import (
    CategoricalHyperparameter, UniformIntHyperparameter,
    UniformFloatHyperparameter, BaseEstimatorHyperparameter)
from ..data_types import AlgorithmType

# TODO: add sgd classifier, xgboost classifier


# ===========
# Naive Bayes
# ===========

def gaussian_naive_bayes():
    """Create a naive bayes algorithm component."""
    return AlgorithmComponent(
        name="GaussianNaiveBayes",
        component_class=GaussianNB,
        component_type=AlgorithmType.CLASSIFIER)


def multinomial_naive_bayes():
    """Create a naive bayes algorithm component."""
    return AlgorithmComponent(
        name="MultinomialNB",
        component_class=MultinomialNB,
        component_type=AlgorithmType.CLASSIFIER,
        hyperparameters=[
            UniformFloatHyperparameter(
                "alpha", 1e-2, 100, default=1, log=True),
            CategoricalHyperparameter(
                "fit_prior", [True, False], default=True)
        ])


# ====================
# Gaussian Classifiers
# ====================


def rbf_gaussian_process_classifier():
    return AlgorithmComponent(
        name="GaussianProcessClassifier",
        component_class=GaussianProcessClassifier,
        component_type=AlgorithmType.CLASSIFIER,
        hyperparameters=[
            UniformFloatHyperparameter(
                "max_iter_predict", 100, 500, default=100)
        ],
        constant_hyperparameters={
            "optimizer": "fmin_l_bfgs_b",
            "n_restarts_optimizer": 10,
        })


# ========================
# Linear Model Classifiers
# ========================

def logistic_regression():
    """Create a logistic regression algorithm component.

    Leave out the following hyperparameters:
    - intercept_scaling
    - random_state
    - warm_start
    - max_iter
    - tol
    """
    return AlgorithmComponent(
        name="LogisticRegression",
        component_class=LogisticRegression,
        component_type=AlgorithmType.CLASSIFIER,
        hyperparameters=[
            CategoricalHyperparameter("penalty", ["l1", "l2"], default="l2"),
            CategoricalHyperparameter(
                "dual", [True, False], default=False),
            CategoricalHyperparameter(
                "fit_intercept", [True, False], default=True),
            CategoricalHyperparameter(
                "class_weight", ["balanced", None], default=None),
            CategoricalHyperparameter(
                "solver", ["newton-cg", "lbfgs", "liblinear", "sag", "saga"],
                default="liblinear"),
            CategoricalHyperparameter(
                "multi_class", ["ovr", "multinomial"], default="ovr"),
            UniformFloatHyperparameter(
                "C", 0.001, 300.0, default=1.0, log=True, n=10),
        ],
        exclusion_conditions={
            "penalty": {
                "l1": {"dual": [True],
                       "solver": ["newton-cg", "lbfgs", "sag"]}},
            "dual": {True: {"solver": ["newton-cg", "lbfgs", "sag", "saga"]}},
            "solver": {"liblinear": {"multi_class": ["multinomial"]}},
        })


# ==========================
# Non-parametric Classifiers
# ==========================

def k_nearest_neighbors():
    """Create K-nearest neighbors classifier component.

    Going to rely on the `algorithm="auto"` and the other default settings,
    following the auto-sklearn implementation:
    https://github.com/automl/auto-sklearn/blob/master/autosklearn/pipeline/components/classification/k_nearest_neighbors.py  # noqa

    The hyperparameters that were left out were:
    - algorithm {'auto', 'ball_tree', 'kd_tree', 'brute'}
    - leaf_size (int)
    - p (int)
    - metric: distance metric to use for tree (default 'minkowski')
    - metric_params: additional keyword arguments for the metric function.
    """
    return AlgorithmComponent(
        name="KNearestNeighorsClassifier",
        component_class=KNeighborsClassifier,
        component_type=AlgorithmType.CLASSIFIER,
        hyperparameters=[
            UniformIntHyperparameter(
                "n_neighbors", 1, 100, log=True, default=1),
            CategoricalHyperparameter(
                "weights", ["uniform", "distance"], default="uniform"),
        ])


# ==========================
# Support-vector Classifiers
# ==========================

def support_vector_classifier_linear():
    """Create linear support vector classifier component.

    hyperparameter settings were taken from:
    https://github.com/automl/auto-sklearn/blob/master/autosklearn/pipeline/components/classification/liblinear_svc.py  # noqa

    See scikit-learn docs for details on constant hyperparameters:
    http://scikit-learn.org/stable/modules/generated/sklearn.svm.LinearSVC.html

    TODO: implement the cache size logic as described in the auto-sklearn
    project https://github.com/automl/auto-sklearn/blob/master/autosklearn/pipeline/components/classification/libsvm_svc.py  # noqa
    """
    return AlgorithmComponent(
        name="LinearSVC",
        component_class=LinearSVC,
        component_type=AlgorithmType.CLASSIFIER,
        hyperparameters=[
            CategoricalHyperparameter("penalty", ["l1", "l2"], default="l2"),
            CategoricalHyperparameter(
                "loss", ["hinge", "squared_hinge"], default="squared_hinge"),
            UniformIntHyperparameter(
                "max_iter", 1000, 3000, default=1000),
            UniformFloatHyperparameter(
                "tol", 1e-5, 1e-1, default=1e-4, log=True),
            UniformFloatHyperparameter(
                "C", 0.03215, 32768, log=True, default=1.0)
        ],
        constant_hyperparameters={
            "multi_class": "ovr",
            "dual": False
        },
        exclusion_conditions={
            "penalty": {
                "l1": {"loss": ["hinge"]},
                "l2": {"loss": ["hinge"]},
            },
        })


def support_vector_classifier_nonlinear():
    """Create nonlinear support vector classifier (rbf, poly, sigmoid)"""
    return AlgorithmComponent(
        name="NonlinearSVC",
        component_class=SVC,
        component_type=AlgorithmType.CLASSIFIER,
        hyperparameters=[
            UniformFloatHyperparameter(
                "C", 0.03125, 32768, default=1.0, log=True),
            CategoricalHyperparameter(
                "kernel", ["rbf", "poly", "sigmoid"], default="rbf"),
            UniformIntHyperparameter("degree", 2, 5, default=3),
            UniformFloatHyperparameter(
                "gamma", 3.0517578125e-05, 8, default=0.1, log=True),
            UniformFloatHyperparameter("coef0", -1, -1, default=0),
            CategoricalHyperparameter(
                "shrinking", [True, False], default=True),
            UniformFloatHyperparameter(
                "tol", 1e-5, 1e-1, default=1e-3, log=True),
        ],
        constant_hyperparameters={
            "max_iter": -1,
            "cache_size": 200,
            "decision_function_shape": "ovr",
        },
        exclusion_conditions={
            "kernel": {
                "rbf": {
                    "degree": EXCLUDE_ALL,
                    "coef0": EXCLUDE_ALL,
                },
                "sigmoid": {"degree": EXCLUDE_ALL}
            },
        })


# ======================
# Tree-based Classifiers
# ======================

def decision_tree():
    """Create a decision tree algorithm component.

    Note here that if the hyperparameter can be either str, int, or float,
    default to float. Future versions will support MultiTypeHyperparameters.

    Leave out the following hyperparameters:
    - min_impurity_split (deprecated)
    - random_state
    - presort (mostly for speed up)
    """
    return AlgorithmComponent(
        name="DecisionTreeClassifier",
        component_class=DecisionTreeClassifier,
        component_type=AlgorithmType.CLASSIFIER,
        hyperparameters=[
            CategoricalHyperparameter(
                "criterion", ["gini", "entropy"], default="gini"),
            CategoricalHyperparameter(
                "splitter", ["best", "random"], default="best"),
            CategoricalHyperparameter(
                "class_weight", ["balanced", None], default=None),
            UniformIntHyperparameter(
                "max_depth", 1, 1000, default=None, log=True),
            UniformFloatHyperparameter(
                "min_samples_split", 0.01, 1.0, default=0.05),
            UniformFloatHyperparameter(
                "min_samples_leaf", 0.01, 0.5, default=0.1),
            UniformFloatHyperparameter(
                "min_weight_fraction_leaf", 0.01, 0.5, default=0.0),
            UniformFloatHyperparameter(
                "max_features", 0.0, 1.0, default=0.33),
            UniformIntHyperparameter(
                "max_leaf_nodes", 2, 1000, default=None, log=True),
            UniformFloatHyperparameter(
                "min_impurity_decrease", 0.0, 1.0, default=0.0, n=10),
        ])


# ====================
# Ensemble Classifiers
# ====================

def adaboost():
    """Create adaboost classifier component."""
    return AlgorithmComponent(
        name="AdaBoostClassifier",
        component_class=AdaBoostClassifier,
        component_type=AlgorithmType.CLASSIFIER,
        hyperparameters=[
            BaseEstimatorHyperparameter(
                hname="base_estimator",
                base_estimator=DecisionTreeClassifier,
                hyperparameters=[
                    UniformIntHyperparameter("max_depth", 1, 10, default=1)
                ],
                default=DecisionTreeClassifier(max_depth=1)
            ),
            UniformIntHyperparameter(
                "n_estimators", 10, 200, default=10, n=10),
            UniformFloatHyperparameter(
                "learning_rate", 0.1, 1., default=1., log=True, n=10),
            CategoricalHyperparameter(
                "algorithm", ["SAMME", "SAMME.R"], default="SAMME.R")
        ])


def extra_trees():
    """Create extra trees classifier component.

    Just follows:
    https://github.com/automl/auto-sklearn/blob/master/autosklearn/pipeline/components/classification/extra_trees.py  # noqa
    """
    return ExtraTreesClassifier(
        name="ExtraTreesClassifier",
        component_class=ExtraTreesClassifier,
        component_type=AlgorithmType.CLASSIFIER,
        hyperparameters=[
            CategoricalHyperparameter(
                "criterion", ["gini", "entropy"], default="gini"),
            UniformFloatHyperparameter(
                "max_features", 0., 1., default=0.5),
            UniformIntHyperparameter("min_samples_split", 2, 20, default=2),
            UniformIntHyperparameter("min_samples_leaf", 1, 20, default=1),
            CategoricalHyperparameter(
                "bootstrap", [True, False], default=False),
            CategoricalHyperparameter(
                "class_weight", ["balanced", "balanced_subsample", None],
                default=None),
        ],
        constant_hyperparameters={
            "n_estimators": 100,
            "max_depth": None,
            "min_weight_fraction_leaf": 0.,
            "max_leaf_nodes": None,
            "min_impurity_decrease": 0.,
        })


def gradient_boosting():
    """Create gradient boosting classifier component."""
    return AlgorithmComponent(
        name="GradientBoostingClassifier",
        component_class=GradientBoostingClassifier,
        component_type=AlgorithmType.CLASSIFIER,
        hyperparameters=[
            UniformFloatHyperparameter(
                "learning_rate", 0.01, 1, default=0.1, log=True),
            UniformIntHyperparameter("n_estimators", 50, 500, default=100),
            UniformIntHyperparameter("max_depth", 1, 10, default=3),
            CategoricalHyperparameter(
                "criterion", ["friedman_mse", "mse", "mae"], default="mse"),
            UniformIntHyperparameter("min_samples_split", 2, 20, default=2),
            UniformIntHyperparameter("min_samples_leaf", 1, 20, default=1),
            UniformFloatHyperparameter("subsample", 0.01, 1.0, default=1.0),
            UniformFloatHyperparameter("max_features", 0.1, 1.0, default=1),
        ],
        constant_hyperparameters={
            "loss": "deviance",
            "min_weight_fraction_leaf": 0,
            "max_leaf_nodes": None,
            "min_impurity_decrease": 0.0,
        })


def random_forest_classifier():
    """Create random forest classifier component."""
    return AlgorithmComponent(
        name="RandomForestClassifier",
        component_class=RandomForestClassifier,
        component_type=AlgorithmType.CLASSIFIER,
        hyperparameters=[
            UniformIntHyperparameter(
                "n_estimators", 50, 200, default=50, n=10),
            CategoricalHyperparameter(
                "criterion", ["gini", "entropy"], default="gini"),
            UniformFloatHyperparameter(
                "max_features", 0.1, 1.0, default=1.0),
            UniformIntHyperparameter(
                "min_samples_split", 2, 20, default=2),
            UniformIntHyperparameter(
                "min_samples_leaf", 1, 20, default=1),
            CategoricalHyperparameter(
                "bootstrap", [True, False], default=True),
            CategoricalHyperparameter(
                "class_weight", ["balanced", None], default=None)
        ],
        constant_hyperparameters={
            "max_depth": None,
            "min_weight_fraction_leaf": 0.0,
            "max_leaf_nodes": None,
            "min_impurity_decrease": 0.0,
        })



"""Logger module."""

import numpy as np


def _metrics():
    return [
        "total_losses",
        "actor_losses",
        "critic_losses",
        "entropy_losses",
        "gradient_norms",
        "mean_rewards",
        "mean_validation_scores",
        "std_validation_scores",
        "n_successful_mlfs",
        "n_unique_mlfs",
        "mlf_diversity",
        "n_unique_hyperparams",
        "hyperparam_diversity",
        "best_validation_scores",
        "entropy_coefficient",
    ]


def empty_logger(*args, **kwargs):
    pass


def _log_floyd(metric, value, step):
    print('{"metric": "%s", "value": %0.10f, "step": %d}' %
          (metric, value, step))


def default_logger(tracker, prefix=""):

    metrics = [
        "total_losses",
        "actor_losses",
        "critic_losses",
        "entropy_losses",
        "gradient_norms",
        "mean_validation_scores",
        "mean_rewards",
        "mlf_diversity"
    ]
    if prefix:
        display_names = ["%s__%s" % (prefix, m) for m in metrics]
    else:
        display_names = metrics
    msg = "\n"
    for n, m in zip(display_names, metrics):
        msg += "%s: %0.02f - " % (n, tracker.history[m][-1])
    print(msg)


def floyd_logger(tracker, prefix=""):
    """Log metrics for floydhub.

    prints metrics to stdout as required by floydhub metrics tracking:
    https://docs.floydhub.com/guides/jobs/metrics/

    :param MetricsTracker tracker: tracker object for deep CASH.
    :param str prefix: prefix metric names with this
    """
    metrics = _metrics()
    metrics_dict = {}
    if prefix:
        metrics_dict = {m: "%s__%s" % (prefix, m) for m in metrics}
    for metric in metrics:
        value = tracker.history[metric][-1]
        step = tracker.history["episode"][-1]
        if value is None or np.isnan(value):
            continue
        _log_floyd(metrics_dict.get(metric, metric), value, step)


def floyd_multiprocess_logger(
        tracker, prefix="", metric="mean_rewards"):
    """Log metrics for floydhub in the case of multiprocessing jobs.

    logs mean_rewards to indicate overall progress of performance.

    :param MetricsTracker tracker: tracker object for deep CASH.
    :param str prefix: prefix metric names with this
    :param str metric: metric to display
    """
    metric_name = prefix + metric if prefix else metric
    if tracker.history[metric] > 0:
        _log_floyd(
            metric_name,
            tracker.history[metric][-1],
            tracker.history["episode"][-1])


def get_loggers():
    """Get dictionary of loggers."""
    return {
        "empty": empty_logger,
        "default": default_logger,
        "floyd": floyd_logger,
        "floyd_multiprocess": floyd_multiprocess_logger,
    }



"""Module for abstracting hyperparameters.

Meant to be a lighter-weight version of https://github.com/automl/ConfigSpace,
which incorporates conditional logic and more features, this API only
enumerates the state space of Hyperparameters without enforcing any rules
about how hyperparameters interact with each other. This will presumably be
the job of the Controller RNN.
"""

from collections import namedtuple
import itertools
import numpy as np

from . import constants

N_VALUES = 5  # number of values to generate for int/float state spaces


class HyperparameterBase(object):

    def __init__(self, hname, state_space, default):
        """Create hyperparameter base class."""
        self.hname = hname
        self._state_space = list(state_space)
        self.default = default

    def __repr__(self):
        return "<%s: \"%s\">" % (type(self).__name__, self.hname)

    def default_in_state_space(self):
        return self.default in self._state_space or self.default is None

    def get_state_space(self, with_none_token=False):
        """
        :returns: list of tokens representing hyperparameter space
        :rtype: list[int|float|str]
        """
        if self.default_in_state_space():
            state_space = self._state_space
        else:
            state_space = self._state_space + [self.default]
        if with_none_token:
            state_space.append(constants.NONE_TOKEN)
        return state_space


class CategoricalHyperparameter(HyperparameterBase):

    def __init__(self, hname, state_space, default):
        super().__init__(hname, state_space, default)


class NumericalHyperparameter(HyperparameterBase):

    def __init__(self, hname, min, max, dtype, default, log, n):
        self.min = dtype(min)
        self.max = dtype(max)
        self.dtype = dtype
        self.n = n
        self.log = log
        super().__init__(hname, self._init_state_space(), default)

    def _init_state_space(self):
        if self.log:
            # evenly distributed in intervals in log space
            space_func = np.geomspace
        else:
            space_func = np.linspace
        return np.sort(
            space_func(self.min, self.max, self.n, dtype=self.dtype))


class UniformIntHyperparameter(NumericalHyperparameter):

    def __init__(self, hname, min, max, default, log=False, n=N_VALUES):
        super().__init__(hname, min, max, int, default, log, n=n)


class UniformFloatHyperparameter(NumericalHyperparameter):

    def __init__(self, hname, min, max, default, log=False, n=N_VALUES):
        super().__init__(hname, min, max, float, default, log, n=n)


class TuplePairHyperparameter(HyperparameterBase):
    """Tuple Pair Hyperparameter class.

    For hyperparameters in the form: `(x, y)`
    """

    def __init__(self, hname, hyperparameters, default):
        self.hyperparameters = hyperparameters
        super().__init__(hname, self._init_state_space(), default)

    def _init_state_space(self):
        return list(
            itertools.product(
                *[h.get_state_space() for h in self.hyperparameters]))


class TupleRepeatingHyperparameter(HyperparameterBase):
    """Tuple Pair Hyperparameter class.

    For hyperparameters in the form: `(x0, x1, ... , xn)`
    """

    def __init__(self, hname, hyperparameter, max_nrepeats, default):
        self.hyperparameter = hyperparameter
        self.max_nrepeats = max_nrepeats
        super().__init__(hname, self._init_state_space(), default)

    def _state_space_n(self, nrepeats):
        return list(
            itertools.product(
                *[self.hyperparameter.get_state_space()
                  for _ in range(nrepeats)]))

    def _init_state_space(self):
        if self.max_nrepeats == 1:
            return [(i,) for i in self.hyperparameter.get_state_space()]
        return list(
            itertools.chain(
                *[self._state_space_n(n)
                  for n in range(1, self.max_nrepeats + 1)]))


class BaseEstimatorHyperparameter(HyperparameterBase):
    """Single Base Estimator Hyperparameter class for ensemble methods.

    For example, for AdaBoost or Bagging estimators.
    """

    def __init__(self, hname, base_estimator, hyperparameters, default):
        """Initialize base estimator hyperparameter."""
        self.base_estimator = base_estimator
        self.hyperparameters = hyperparameters
        super().__init__(hname, self._init_state_space(), default)

    def default_in_state_space(self):
        for base_est in self._state_space:
            if self.default.get_params() == base_est.get_params():
                return True
        return False

    def _init_state_space(self):
        # TODO: similar implementation of this in
        # AlgorithmComponent.hyperparameter_iterator. Consider abstracting that
        # functionality into a utils module.
        expanded_state_space = []
        for h in self.hyperparameters:
            expanded_state_space.append([
                (h.hname, v) for v in h.get_state_space()])
        return [
            self.base_estimator(**dict(hsetting)) for hsetting in
            list(itertools.product(*expanded_state_space))]


class EmbeddedEstimatorHyperparameter(HyperparameterBase):
    """Sets the hyperparameter value of an embedded estimator.

    For example, in the ColumnTransformer ``transformer`` argument.
    """

    def __init__(self, estimator_name, hyperparameter):
        self.hyperparameter = hyperparameter
        self.estimator_name = estimator_name
        super().__init__(
            f"{self.estimator_name}__{self.hyperparameter.hname}",
            hyperparameter.get_state_space(),
            hyperparameter.default)


class MultiBaseEstimatorHyperparameter(HyperparameterBase):
    """Multiple Base Estimator Hyperparameter class for ensemble methods.

    TODO: this should take a list of BaseEstimatorHyperparameters and the
    state space is the concatenation of all possible estimators.
    """
    pass


class MultiTypeHyperparameters(HyperparameterBase):
    """TODO: This should take a list of Hyperparameters.

    Support a hyperparameter that can be multiple types.
    """
    pass


class ProbabalisticHyperparameterBase(object):
    """
    TODO: This should be a family of hyperparameters that are numerical and
    can be drawn from a distribution, e.g. Gaussian.
    """
    pass



"""Reinforce module for training the CASH controller."""

from collections import namedtuple

import numpy as np
import torch
import torch.nn as nn

from . import utils, loggers
from .tracking import MetricsTracker

EPSILON = np.finfo(np.float32).eps.item()
CLIP_GRAD = 1.0

Losses = namedtuple(
    "Losses", ["total", "actor", "critic", "entropy", "grad_norm"])


class MetaLearnReinforce(object):
    """Reinforce component of deep-cash algorithm."""

    def __init__(
            self,
            controller,
            t_env,
            gamma=0.99,
            entropy_coef=0.0,
            entropy_coef_anneal_to=0.0,
            entropy_coef_anneal_by=None,
            normalize_reward=False,
            meta_reward_multiplier=1.,
            sample_new_task_every=10,
            metrics_logger=loggers.default_logger):
        """Initialize CASH Reinforce Algorithm.

        :param pytorch.nn.Module controller: A CASH controller to select
            actions.
        :param TaskEnvironment t_env: task environment to sample data
            environments and evaluate proposed ml frameworks.
        :param float gamma: discounted reward for computing returns.
        :param float entropy_coef: coefficient for entropy regularization.
        :param float entropy_coef: coefficient to anneal to by end of training.
        :param float entropy_coef_anneal_by: Value between 0.0 and 1.0
            indicating when during the course of n_episodes the entropy
            coefficient should reach 0. If None, no annealing is performed.
            For example if this value is 0.5, then the entropy coefficient will
            linearly reach 0.0 half-way through training.
        :param bool normalize_reward: whether or not to mean-center and
            standard-deviation-scale the reward signal for backprop.
        :param float meta_reward_multiplier: multiply the reward signal with
            this value. If this is set to 0, this effectively disables the
            metalearning reward signal.
        :param callable metrics_logger: logging function to use. The function
            takes as input a tracking.TrackerBase object and prints out a
            message.
        """
        self.controller = controller
        self.t_env = t_env
        self._gamma = gamma
        self._entropy_coef = entropy_coef
        self._entropy_coef_anneal_to = entropy_coef_anneal_to
        self._entropy_coef_anneal_by = entropy_coef_anneal_by
        self._normalize_reward = normalize_reward
        self._meta_reward_multiplier = meta_reward_multiplier
        self._metrics_logger = metrics_logger

        if entropy_coef_anneal_by is not None and \
                not 0 < entropy_coef_anneal_by <= 1:
            raise ValueError(
                "entropy_coef_anneal must be between (0, 1], found %s" %
                entropy_coef_anneal_by)

    def fit(self, optim, optim_kwargs, n_episodes=100, n_iter=100,
            verbose=True, procnum=None):
        """Fits the CASH controller with the REINFORCE algorithm.

        :param torch.optim optim: type of optimization to use for backprop.
        :param dict optim_kwargs: arguments to pass to optim constructor
        :param int n_episodes: number of episodes to train.
        :param int n_iter: number of iterations per episode.
        :param bool verbose: whether or not to print the exponential mean
            reward per iteration.
        :param int procnum: optional argument to indicate the process number
            used for multiprocessing.
        """
        # set optimization object
        self.optim = optim(self.controller.parameters(), **optim_kwargs)
        self._n_episodes = n_episodes

        # set entropy coefficient schedule per episode
        if self._entropy_coef_anneal_by is not None:
            episodes_to_zero = int(n_episodes * self._entropy_coef_anneal_by)
            # linearly decrease coefficient to zero
            self._entropy_coef_schedule = np.linspace(
                self._entropy_coef,
                self._entropy_coef_anneal_to,
                num=episodes_to_zero).tolist()
            self._entropy_coef_schedule += [
                self._entropy_coef_anneal_to
                for _ in range(n_episodes - episodes_to_zero)]
        else:
            self._entropy_coef_schedule = [self._entropy_coef] * n_episodes

        # track metrics
        self.tracker = MetricsTracker()

        for i_episode in range(self._n_episodes):
            self.t_env.sample_data_env()
            self._action_buffer = []
            self._validation_scores = []
            self._successful_mlfs = []
            self._algorithm_sets = []
            self._hyperparameter_sets = []
            self._best_validation_score = None
            self._best_mlf = None
            msg = "episode %d, task: %s" % (
                i_episode + 1, self.t_env.current_data_env.name)
            if procnum is not None:
                msg = "proc num: %d, %s" % (procnum, msg)
            print("\n" + msg)

            last_value = self.start_episode(n_iter, verbose)

            # accumulate stats
            n_successful_mlfs = len(self._successful_mlfs)
            n_unique_mlfs = len(set((tuple(s) for s in self._algorithm_sets)))
            n_unique_hyperparams = len(set(
                [str(d.items()) for d in self._hyperparameter_sets]))
            mlf_diversity = utils._diversity_metric(
                n_unique_mlfs, n_successful_mlfs)
            hyperparam_diversity = utils._diversity_metric(
                n_unique_hyperparams, n_successful_mlfs)

            mean_rewards = np.mean(self.controller.reward_buffer)
            if len(self._validation_scores) > 0:
                mean_val_scores = np.mean(self._validation_scores)
                std_val_scores = np.std(self._validation_scores)
            else:
                mean_val_scores, std_val_scores = np.nan, np.nan

            # backward pass through controller
            losses = self.end_episode(
                last_value, self._entropy_coef_schedule[i_episode])

            self.tracker.update_metrics({
                "episode": i_episode + 1,
                "data_env_names": self.t_env.current_data_env.name,
                "target_type": self.t_env.current_data_env.target_type.name,
                "scorers": self.t_env.scorer.name,
                "total_losses": losses.total,
                "actor_losses": losses.actor,
                "critic_losses": losses.critic,
                "entropy_losses": losses.entropy,
                "gradient_norms": losses.grad_norm,
                "mean_rewards": mean_rewards,
                "mean_validation_scores": mean_val_scores,
                "std_validation_scores": std_val_scores,
                "best_validation_scores": self._best_validation_score,
                "best_mlfs": self._best_mlf,
                "n_successful_mlfs": n_successful_mlfs,
                "n_unique_mlfs": n_unique_mlfs,
                "n_unique_hyperparams": n_unique_hyperparams,
                "mlf_diversity": mlf_diversity,
                "hyperparam_diversity": hyperparam_diversity,
                "entropy_coefficient": self._entropy_coef_schedule[i_episode],
            })
            if self._metrics_logger is not None:
                self._metrics_logger(self.tracker)
        return self

    @property
    def history(self):
        """Get metrics history per episode."""
        return self.tracker.history

    @property
    def best_mlfs(self):
        """Get best mlfs per episode."""
        return self.tracker.best_mlfs

    def start_episode(self, n_iter, verbose):
        """Begin training the controller."""
        self._n_valid_mlf = 0
        prev_reward, prev_action = 0, self.controller.init_action()
        prev_hidden = self.controller.init_hidden()
        for i in range(n_iter):
            state = self.t_env.sample_task_state()
            prev_reward, prev_action, prev_hidden = self._fit_iter(
                state,
                self.t_env.current_data_env.target_type,
                prev_reward * self._meta_reward_multiplier,
                prev_action,
                prev_hidden)
            if verbose:
                print(
                    "iter %d - n valid mlf: %d/%d" % (
                        i, self._n_valid_mlf, i + 1),
                    sep=" ", end="\r", flush=True)

        state = self.t_env.sample_task_state()
        last_value, *_ = self.controller(
            prev_action=prev_action,
            prev_reward=utils.aux_tensor(prev_reward),
            metafeatures=state,
            hidden=prev_hidden,
            target_type=self.t_env.current_data_env.target_type,
        )
        return last_value.detach().item()

    def _fit_iter(
            self,
            metafeature_tensor,
            target_type,
            prev_reward,
            prev_action,
            prev_hidden):
        value, actions, action_activation, hidden = self.controller(
            prev_action=prev_action,
            prev_reward=utils.aux_tensor(prev_reward),
            metafeatures=metafeature_tensor,
            hidden=prev_hidden,
            target_type=target_type,
        )
        reward = self.evaluate_actions(actions, action_activation)

        self.controller.value_buffer.append(value)
        self.controller.log_prob_buffer.append(
            [a["log_prob"] for a in actions])
        self.controller.reward_buffer.append(reward)
        self.controller.entropy_buffer.append([a["entropy"] for a in actions])

        return reward, action_activation, hidden

    def evaluate_actions(self, actions, action_activation):
        """Evaluate actions on the validation set of the data environment."""
        algorithms, hyperparameters = utils.get_mlf_components(actions)
        mlf = self.controller.a_space.create_ml_framework(
            algorithms, hyperparameters=hyperparameters,
            task_metadata=self.t_env.get_current_task_metadata())
        mlf, reward, score = self.t_env.evaluate(mlf)
        self._action_buffer.append(action_activation)
        if reward is None:
            return self.t_env.error_reward
        else:
            self._validation_scores.append(score)
            self._n_valid_mlf += 1
            self._successful_mlfs.append(mlf)
            self._algorithm_sets.append(algorithms)
            self._hyperparameter_sets.append(hyperparameters)
            if self._best_validation_score is None or \
                    self.t_env.scorer.comparator(
                        score, self._best_validation_score):
                self._best_validation_score = score
                self._best_mlf = mlf
            return reward

    def end_episode(self, last_value, entropy_coef):
        """End an episode with one backpropagation step.

        Since REINFORCE algorithm is a gradient ascent method, negate the log
        probability in order to do gradient descent on the negative expected
        rewards.

        :returns: tuple of loss and aggregated gradient.
        :rtype: tuple[float, float]
        """
        _check_buffers(
            self.controller.value_buffer,
            self.controller.log_prob_buffer,
            self.controller.reward_buffer,
            self.controller.entropy_buffer)

        n = len(self.controller.reward_buffer)

        # reset gradient
        self.optim.zero_grad()

        # compute Q values
        returns = torch.zeros(len(self.controller.value_buffer))
        R = last_value
        for t in reversed(range(n)):
            R = self.controller.reward_buffer[t] + self._gamma * R
            returns[t] = R

        # mean-center and std-scale returns
        if self._normalize_reward:
            returns = (returns - returns.mean()) / (returns.std() + EPSILON)

        values = torch.cat(self.controller.value_buffer).squeeze()
        advantage = returns - values

        # compute loss
        actor_loss = [
            -l * a
            for log_probs, a in zip(self.controller.log_prob_buffer, advantage)
            for l in log_probs
        ]
        actor_loss = torch.cat(actor_loss).mean()
        critic_loss = 0.5 * advantage.pow(2).mean()

        # entropy loss term, negate the mean since we want to maximize entropy
        entropies = torch.cat([
            e * entropy_coef for entropy_list in self.controller.entropy_buffer
            for e in entropy_list
        ]).squeeze()
        entropy_loss = -entropies.mean()

        actor_critic_loss = actor_loss + critic_loss + entropy_loss

        # one step of gradient descent
        actor_critic_loss.backward()
        # gradient clipping to prevent exploding gradient
        nn.utils.clip_grad_norm_(self.controller.parameters(), CLIP_GRAD)
        self.optim.step()

        grad_norm = 0.
        for p in self.controller.parameters():
            if p.grad is not None:
                param_norm = p.grad.data.norm(2)
                grad_norm += param_norm.item() ** 2
        grad_norm = grad_norm ** 0.5

        # reset rewards and log probs
        del self.controller.value_buffer[:]
        del self.controller.log_prob_buffer[:]
        del self.controller.reward_buffer[:]
        del self.controller.entropy_buffer[:]

        return Losses(
            actor_critic_loss.data.item(),
            actor_loss.data.item(),
            critic_loss.data.item(),
            entropy_loss.data.item(),
            grad_norm)


def _check_buffers(
        value_buffer, log_prob_buffer, reward_buffer, entropy_buffer):
    n_vbuf = len(value_buffer)
    n_abuf = len(log_prob_buffer)
    n_rbuf = len(reward_buffer)
    n_ebuf = len(entropy_buffer)
    if n_vbuf != n_abuf != n_rbuf != n_ebuf:
        raise ValueError(
            "all buffers need the same length, found:\n"
            "value_buffer = %d\n"
            "log_prob_buffer = %d\n"
            "reward_buffer = %d\n"
            "entropy_buffer = %d\n" % (n_abuf, n_rbuf, n_ebuf))



"""Module for sampling from a distribution of classification environments."""

from collections import OrderedDict
from functools import partial

import sklearn.datasets
import sklearn.preprocessing

from .data_environment import DataEnvironment
from ..data_types import FeatureType, TargetType, DataSourceType


def envs(n=5, test_size=None, random_state=None, verbose=False):
    # TODO: limit number of envs by n
    return OrderedDict([
        ("sklearn.iris", DataEnvironment(
            name="sklearn.iris",
            source=DataSourceType.SKLEARN,
            target_type=TargetType.MULTICLASS,
            feature_types=[FeatureType.CONTINUOUS for _ in range(4)],
            feature_indices=range(4),
            fetch_training_data=partial(
                sklearn.datasets.load_iris, return_X_y=True),
            fetch_test_data=None,
            test_size=test_size,
            random_state=random_state,
            target_preprocessor=None,
            scorer=None)),
        ("sklearn.digits", DataEnvironment(
            name="sklearn.digits",
            source=DataSourceType.SKLEARN,
            target_type=TargetType.MULTICLASS,
            feature_types=[FeatureType.CONTINUOUS for _ in range(64)],
            feature_indices=range(64),
            fetch_training_data=partial(
                sklearn.datasets.load_digits, return_X_y=True),
            fetch_test_data=None,
            test_size=test_size,
            random_state=random_state,
            target_preprocessor=None,
            scorer=None)),
        ("sklearn.wine", DataEnvironment(
            name="sklearn.wine",
            source=DataSourceType.SKLEARN,
            target_type=TargetType.MULTICLASS,
            feature_types=[FeatureType.CONTINUOUS for _ in range(13)],
            feature_indices=range(13),
            fetch_training_data=partial(
                sklearn.datasets.load_wine, return_X_y=True),
            fetch_test_data=None,
            test_size=test_size,
            random_state=random_state,
            target_preprocessor=None,
            scorer=None)),
        ("sklearn.breast_cancer", DataEnvironment(
            name="sklearn.breast_cancer",
            source=DataSourceType.SKLEARN,
            target_type=TargetType.BINARY,
            feature_types=[FeatureType.CONTINUOUS for _ in range(30)],
            feature_indices=range(39),
            fetch_training_data=partial(
                sklearn.datasets.load_breast_cancer, return_X_y=True),
            fetch_test_data=None,
            test_size=test_size,
            random_state=random_state,
            target_preprocessor=None,
            scorer=None)),
    ])



"""Module for creating, reading deep cash experiments."""

import datetime
import inspect
import itertools
import joblib
import logging
import os
import pandas as pd
import subprocess
import torch
import torch.multiprocessing as mp
import yaml
import yamlordereddictloader

from collections import namedtuple, OrderedDict
from functools import partial
from pathlib import Path

from .algorithm_space import AlgorithmSpace
from .evaluate import evaluate_controller
from .task_environment import TaskEnvironment
from .metalearn_controller import MetaLearnController
from .metalearn_reinforce import MetaLearnReinforce
from .data_environments.environments import envs
from .data_types import ExperimentType
from .loggers import get_loggers, empty_logger
from . import utils

# TODO: support ability to copy a config file with new name
# TODO: support ability to run floyd job by experiment name


logger = logging.getLogger(__name__)


ExperimentConfig = namedtuple(
    "ExperimentConfig", [
        "name",
        "experiment_type",
        "description",
        "created_at",
        "git_hash",
        "parameters",
        "notes",
    ])


def _env_names():
    return [d["dataset_name"] for d in envs()]


def get_experiment_fn(experiment_type):
    return {
        ExperimentType.METALEARN_REINFORCE: run_experiment,
    }[ExperimentType[experiment_type]]


def get_default_parameters(experiment_type):
    experiment_fn = get_experiment_fn(experiment_type)
    return OrderedDict([
        (k, v.default) for k, v in
        inspect.signature(experiment_fn).parameters.items()
    ])


def create_config(name, experiment_type, description=None,
                  date_format="%Y-%m-%d-%H-%M-%S", **custom_parameters):
    parameters = get_default_parameters(experiment_type)
    if custom_parameters:
        parameters.update(custom_parameters)
    created_at = datetime.datetime.now()
    git_hash = subprocess.check_output(["git", "rev-parse", "--short", "HEAD"])
    return ExperimentConfig( 
        name=name,
        experiment_type=experiment_type,
        description=description,
        created_at=created_at.strftime(date_format),
        git_hash=git_hash.strip().decode("utf-8"),
        parameters=parameters,
        notes=None,
    )


def write_config(config, dir_path, fname=None):
    dir_path = Path(dir_path)
    dir_path.mkdir(exist_ok=True)
    fp = _config_filepath(dir_path, config) if fname is None \
        else dir_path / fname
    with fp.open("w") as f:
        yaml.dump(config._asdict(), f, Dumper=yamlordereddictloader.Dumper,
                  default_flow_style=False)
    print("wrote experiment config file to %s" % fp)


def read_config(fp):
    with open(fp, "r") as f:
        config_dict = yaml.load(f, Loader=yamlordereddictloader.Loader)
    if "notes" not in config_dict:
        config_dict["notes"] = None
    return ExperimentConfig(**config_dict)


def _config_filepath(dir_path, config):
    return dir_path / (
        "experiment_%s_%s.yml" % (config.created_at, config.name))


def save_best_mlfs(data_path, best_mlfs, proc_num):
    mlf_path = data_path / ("metalearn_controller_mlfs_trial_%d" % proc_num)
    mlf_path.mkdir(exist_ok=True)
    for i, mlf in enumerate(best_mlfs):
        joblib.dump(
            mlf, mlf_path / ("best_mlf_episode_%d.pkl.gz" % (i + 1)),
            compress=3,
        )


def create_hyperparameter_grid(hyperparameters):
    allowable_hyperparameters = [
        "input_size",
        "hidden_size",
        "output_size",
        "n_layers",
        "n_episodes",
        "entropy_coef",
        "entropy_coef_anneal_to",
        "learning_rate",
        "meta_reward_multiplier",
        "normalize_reward",
        "error_reward",
    ]

    if isinstance(hyperparameters, dict):
        hyperparameters = {
            k: v for k, v in hyperparameters.items() if v is not None
        }
        hyperparameter_grid = itertools.product(*[
            [(hyperparameter, value)
             for value in hyperparameters[hyperparameter]]
            for hyperparameter in allowable_hyperparameters
            if hyperparameter in hyperparameters
        ])
        return list(map(dict, hyperparameter_grid))
    elif isinstance(hyperparameters, list):
        return [
            {k: v for k, v in h_dict.items() if k in allowable_hyperparameters}
            for h_dict in hyperparameters
        ]
    else:
        raise ValueError(
            "hyperparameters type %s not recognized" % type(hyperparameters)
        )


def run_experiment(
        datasets=None,
        test_datasets=None,
        output_fp=os.path.dirname(__file__) + "/../output",
        input_size=30,
        hidden_size=30,
        output_size=30,
        n_layers=3,
        dropout_rate=0.2,
        entropy_coef=0.0,
        entropy_coef_anneal_to=0.0,
        entropy_coef_anneal_by=None,
        normalize_reward=True,
        gamma=0.99,
        meta_reward_multiplier=1.,
        n_episodes=100,
        n_iter=16,
        n_eval_iter=10,
        n_eval_samples=5,
        learning_rate=0.005,
        optim_beta1=0.9,
        optim_beta2=0.999,
        env_sources=["SKLEARN", "OPEN_ML", "KAGGLE"],
        test_env_sources=["AUTOSKLEARN_BENCHMARK"],
        target_types=["BINARY", "MULTICLASS"],
        test_env_target_types=["BINARY", "MULTICLASS"],
        test_set_config={
            "SKLEARN": {
                "test_size": 0.8,
                "random_state": 100,
            },
            "AUTOSKLEARN_BENCHMARK": {
                "test_size": 0.8,
                "random_state": 100,
            },
        },
        error_reward=0,
        n_samples=5000,
        per_framework_time_limit=180,
        per_framework_memory_limit=5000,
        metric_logger=None,
        fit_verbose=1,
        controller_seed=1000,
        task_environment_seed=100,
        hyperparameters=None):
    """Run deep cash experiment with single configuration."""
    torch.manual_seed(controller_seed)
    output_fp = os.path.dirname(__file__) + "/../output" if \
        output_fp is None else output_fp
    data_path = Path(output_fp)
    data_path.mkdir(exist_ok=True)
    hyperparameters = {} if hyperparameters is None else hyperparameters

    # initialize error logging (this is to log fit/predict/score errors made
    # when evaluating a proposed MLF)
    utils.init_logging(str(Path(output_fp) / "fit_predict_error_logs.log"))

    # this logger is for logging metrics to floydhub/stdout
    metric_logger = get_loggers().get(metric_logger, empty_logger)

    def fit_metalearn(proc_num, reinforce, **hyperparameters):
        """Fit Metalearning Algorithm helper."""
        reinforce.t_env.reset_random_state()
        reinforce.fit(
            optim=torch.optim.Adam,
            optim_kwargs={
                "lr": hyperparameters.get("learning_rate", learning_rate),
                "betas": (optim_beta1, optim_beta2),
            },
            n_episodes=hyperparameters.get("n_episodes", n_episodes),
            n_iter=n_iter,
            verbose=bool(int(fit_verbose)),
            procnum=proc_num)

        # serialize reinforce controller
        reinforce.controller.save(
            data_path / f"controller_trial_{proc_num}.pt")

        # serialize best mlfs
        save_best_mlfs(data_path, reinforce.best_mlfs, proc_num)

        evaluation_results = evaluate_controller(
            reinforce.controller,
            reinforce.t_env,
            meta_reward_multiplier=hyperparameters.get(
                "meta_reward_multiplier", meta_reward_multiplier),
            n_shots=n_eval_iter,
            n_eval_samples=n_eval_samples,
        )

        for env_key, results in evaluation_results.items():
            if results is None:
                continue
            fname = f"{env_key}_inference_results_trial_{proc_num}.csv"
            with (data_path / fname).open("w") as f:
                results.to_csv(f, index=False)

        history = pd.DataFrame(reinforce.history).assign(
            trial_num=proc_num, **hyperparameters)
        history.to_csv(
            data_path / f"metalearn_training_results_trial_{proc_num}.csv",
            index=False)

    processes = []
    hyperparameter_grid = create_hyperparameter_grid(hyperparameters)
    print("hyperparameter grid: %s" % hyperparameter_grid)
    for proc_num, _hyperparameters in enumerate(hyperparameter_grid):
        t_env = TaskEnvironment(
            env_sources=env_sources,
            test_env_sources=test_env_sources,
            target_types=target_types,
            test_env_target_types=test_env_target_types,
            test_set_config=test_set_config,
            random_state=task_environment_seed,
            per_framework_time_limit=per_framework_time_limit,
            per_framework_memory_limit=per_framework_memory_limit,
            dataset_names=datasets,
            test_dataset_names=test_datasets,
            error_reward=_hyperparameters.get("error_reward", error_reward),
            n_samples=n_samples)

        a_space = AlgorithmSpace(
            with_end_token=False,
            hyperparam_with_start_token=False,
            hyperparam_with_none_token=False)

        controller = MetaLearnController(
            metafeature_size=t_env.metafeature_dim,
            input_size=_hyperparameters.get("input_size", input_size),
            hidden_size=_hyperparameters.get("hidden_size", hidden_size),
            output_size=_hyperparameters.get("output_size", output_size),
            a_space=a_space,
            dropout_rate=dropout_rate,
            num_rnn_layers=_hyperparameters.get("n_layers", n_layers),
        )

        reinforce = MetaLearnReinforce(
            controller,
            t_env,
            gamma=gamma,
            entropy_coef=_hyperparameters.get("entropy_coef", entropy_coef),
            entropy_coef_anneal_to=_hyperparameters.get(
                "entropy_coef_anneal_to", entropy_coef_anneal_to),
            entropy_coef_anneal_by=entropy_coef_anneal_by,
            normalize_reward=_hyperparameters.get(
                "normalize_reward", normalize_reward),
            meta_reward_multiplier=_hyperparameters.get(
                "meta_reward_multiplier", meta_reward_multiplier),
            metrics_logger=partial(
                metric_logger, prefix="proc_num_%d" % proc_num)
        )

        p = mp.Process(
            target=fit_metalearn,
            args=(proc_num, reinforce),
            kwargs=_hyperparameters,
        )
        p.start()
        processes.append(p)

    for p in processes:
        p.join()

    for i in range(len(hyperparameter_grid)):
        print("loading controllers from each trial")
        try:
            controller = reinforce.controller.load(
                data_path / ("controller_trial_%d.pt" % i))
            print(controller)
        except Exception:
            print("could not read controller for process %d" % i)



"""Modules implementing random search."""

import numpy as np

from . import utils, loggers
from .algorithm_space import TARGET_TYPE_TO_MLF_SIGNATURE
from .tracking import MetricsTracker


class CASHRandomSearch(object):

    def __init__(self, a_space, t_env, metrics_logger=loggers.default_logger):
        """Initialize random search ML framework fitter."""
        self.a_space = a_space
        self.t_env = t_env
        self._metrics_logger = metrics_logger

    def fit(self, n_episodes: int, n_iter: int, verbose: bool = False,
            procnum: int = None):
        print("Running Random search CASH procedure.")
        self.tracker = MetricsTracker()
        for i_episode in range(n_episodes):
            # TODO: not sure if this for loop should be parallelized
            self.t_env.sample_data_env()
            msg = "episode %d, task: %s" % (
                i_episode, self.t_env.current_data_env.name)
            if procnum is not None:
                msg = "proc num: %d, %s" % (procnum, msg)
            print("\n" + msg)
            self.end_episode(
                i_episode, n_iter, *self.start_episode(n_iter, verbose))
            if self._metrics_logger is not None:
                self._metrics_logger(self.tracker)

    def start_episode(self, n_iter, verbose):
        mlf_signature = TARGET_TYPE_TO_MLF_SIGNATURE[
            self.t_env.current_data_env.target_type]
        mlfs, rewards, scores = [], [], []
        for i in range(n_iter):
            # TODO: this for loop can be parallelized
            self.t_env.sample_task_state()
            mlf = self.a_space.sample_ml_framework(mlf_signature)
            mlf, reward, score = self.t_env.evaluate(mlf)
            if reward is None:
                reward = self.t_env.error_reward
            else:
                mlfs.append(mlf)
                scores.append(score)
            rewards.append(reward)
            if verbose:
                print(
                    "iter %d - n valid mlf: %d/%d" % (
                        i, len(mlfs), i + 1),
                    sep=" ", end="\r", flush=True)
        return mlfs, rewards, scores

    def end_episode(self, i_episode, n_iter, mlfs, rewards, scores):
        # accumulate stats
        n_successful_mlfs = len(mlfs)
        n_unique_mlfs = len(set(
            [tuple(mlf.named_steps.keys()) for mlf in mlfs]))
        n_unique_hyperparams = len(set(
            [str(mlf.get_params().items()) for mlf in mlfs]))
        mlf_diversity = utils._diversity_metric(
            n_unique_mlfs, n_successful_mlfs)
        hyperparam_diversity = utils._diversity_metric(
            n_unique_hyperparams, n_successful_mlfs)

        # NOTE that the random search fitter doesn't use rewards but it's
        # collect in order to compare with CASH agent.
        mean_rewards = np.mean(rewards)
        if len(scores) > 0:
            mean_val_scores = np.mean(scores)
            std_val_scores = np.std(scores)
            best_validation_score = np.max(scores)
            best_mlf = mlfs[np.argmax(scores)]
        else:
            mean_val_scores, std_val_scores = np.nan, np.nan
            best_validation_score, best_mlf = np.nan, np.nan

        self.tracker.update_metrics({
            "episode": i_episode + 1,
            "data_env_names": self.t_env.current_data_env.name,
            "scorers": self.t_env.scorer.name,
            "total_losses": np.nan,
            "actor_losses": np.nan,
            "critic_losses": np.nan,
            "entropy_losses": np.nan,
            "mean_rewards": mean_rewards,
            "gradient_norms": np.nan,
            "mean_validation_scores": mean_val_scores,
            "std_validation_scores": std_val_scores,
            "best_validation_scores": best_validation_score,
            "best_mlfs": best_mlf,
            "n_successful_mlfs": n_successful_mlfs,
            "n_unique_mlfs": n_unique_mlfs,
            "n_unique_hyperparams": n_unique_hyperparams,
            "mlf_diversity": mlf_diversity,
            "hyperparam_diversity": hyperparam_diversity,
            "entropy_coefficient": np.nan,
        })
        if self._metrics_logger is not None:
            self._metrics_logger(self.tracker)

    @property
    def history(self):
        return self.tracker.history

    @property
    def best_mlfs(self):
        return self.tracker.best_mlfs



"""Metadata for kaggle classification datasets."""

import itertools

from collections import OrderedDict

from .kaggle_base import KaggleCompetition

from ..data_types import FeatureType, TargetType
from .. import scorers

from .feature_maps import (
    kaggle_costa_rican_household_poverty_prediction,
    kaggle_homesite_quote_conversion,
    kaggle_santander_customer_satisfaction,
)


def homesite_quote_conversion():
    """Create data interface to kaggle 'homesite quote conversion'.

    url: https://www.kaggle.com/c/homesite-quote-conversion
    """
    features = kaggle_homesite_quote_conversion.get_feature_map()

    def _preprocessor(df):
        """Deal with quirk in the dataset where floats have commas."""
        for f, ftype in features.items():
            if ftype is FeatureType.CONTINUOUS:
                if df[f].dtype == "object":
                    x = df[f].str.replace(",", "")
                    df[f] = x.astype(float)
        return df

    return KaggleCompetition(
        competition_id="homesite-quote-conversion",
        features=features,
        target={"QuoteConversion_Flag": TargetType.BINARY},
        training_data_fname="train.csv.zip",
        test_data_fname="test.csv.zip",
        custom_preprocessor=_preprocessor,
        scorer=scorers.roc_auc())


def santander_customer_satisfaction():
    """Create data interface to kaggle 'santander customer satisfaction'.

    url: https://www.kaggle.com/c/santander-customer-satisfaction
    """
    return KaggleCompetition(
        competition_id="santander-customer-satisfaction",
        features=kaggle_santander_customer_satisfaction.get_feature_map(),
        target={"TARGET": TargetType.BINARY},
        training_data_fname="train.csv.zip",
        test_data_fname="test.csv.zip",
        custom_preprocessor=None,
        scorer=scorers.roc_auc())


def bnp_paribas_cardif_claims_management():
    """Create data interface to kaggle 'bnp paribas cardif claims management'.

    url: https://www.kaggle.com/c/bnp-paribas-cardif-claims-management
    """
    categorical_features = [
        "v%d" % i for i in itertools.chain(
            [3, 22, 24],
            [30, 31, 47, 52, 56, 66, 71, 74, 75, 79, 91, 107, 110, 112, 113,
             125])]
    continuous_features = [
        "v%d" % i for i in itertools.chain(
            [1, 2],
            range(4, 22),
            [23],
            range(25, 30),
            range(32, 47),
            range(48, 52),
            range(53, 56),
            range(57, 66),
            range(67, 71),
            [72, 73, 76, 77, 78],
            range(80, 91),
            range(92, 107),
            [108, 109, 111],
            range(114, 125),
            range(126, 132))]

    return KaggleCompetition(
        competition_id="bnp-paribas-cardif-claims-management",
        features=OrderedDict(
            itertools.chain(
                ((f, FeatureType.CATEGORICAL) for f in categorical_features),
                ((f, FeatureType.CONTINUOUS) for f in continuous_features)
            )),
        target={"target": TargetType.BINARY},
        training_data_fname="train.csv.zip",
        test_data_fname="test.csv.zip",
        custom_preprocessor=None,
        scorer=scorers.log_loss())


def poker_rule_induction():
    """Create data interface to kaggle 'poker rule induction'.

    url: https://www.kaggle.com/c/poker-rule-induction
    """
    letters = [
        "S",  # suit: 1 - 13 (Ace, 2, 3, ..., King)
        "C",  # card: 1 - 4 (Hearts, Spades, Diamonds, Clubs)
    ]
    return KaggleCompetition(
        competition_id="poker-rule-induction",
        features=OrderedDict([
            ("%s%d" % (i, j), FeatureType.CONTINUOUS)
            for i in letters for j in range(1, 6)
        ]),
        target={"hand": TargetType.MULTICLASS},
        training_data_fname="train.csv.zip",
        test_data_fname="test.csv.zip",
        custom_preprocessor=None,
        scorer=scorers.accuracy())


def costa_rican_household_poverty_prediction():
    """Create data interface to kaggle 'costa rican household poverty prediction'.

    url: https://www.kaggle.com/c/costa-rican-household-poverty-prediction
    """
    def _preprocessor(df):
        """Deal with quirks in the costa rican household poverty dataset.

        https://www.kaggle.com/c/costa-rican-household-poverty-prediction/data

        This function deals with a quirk in the dataset where "yes" and "no"
        values are present in the following continuous columns:
        - dependency
        - edjefe
        - edjefa

        "yes" values are meant to be 1, and "no" values are meant to be 0
        """
        replace_dict = {"yes": "1", "no": "0"}
        return (
            df.replace({
                "dependency": replace_dict,
                "edjefe": replace_dict,
                "edjefa": replace_dict,
            })
            .astype({
                "dependency": float,
                "edjefe": float,
                "edjefa": float
            })
        )

    return KaggleCompetition(
        competition_id="costa-rican-household-poverty-prediction",
        features=kaggle_costa_rican_household_poverty_prediction.get_feature_map(),  # noqa E501
        target={"Target": TargetType.MULTICLASS},
        training_data_fname="train.csv.zip",
        test_data_fname="test.csv.zip",
        custom_preprocessor=_preprocessor,
        scorer=scorers.f1_score_macro())



"""Module for generating algorithm and hyperparameter sequences.

This is a controller that jointly generates machine learning framework
pipelines and their associated hyperparameter settings. It does so via an RNN
that has a separate softmax classifier at each time step, where the classifier
has to predict either an algorithm to add to the pipeline or a hyperparameter
for the most recently chosen algorithm.
"""

from collections import defaultdict

import dill
import logging
import torch
import torch.nn as nn
import torch.nn.functional as F

from torch.distributions import Categorical

from .data_types import CASHComponent, TargetType
from .components.algorithm_component import EXCLUDE_ALL


logger = logging.getLogger(__name__)


class MetaLearnController(nn.Module):
    """RNN module to generate joint algorithm and hyperparameter controller."""

    def __init__(
            self,
            metafeature_size,
            input_size,
            hidden_size,
            output_size,
            a_space,
            mlf_signature=None,
            metafeature_encoding_size=20,
            aux_reward_size=1,
            dropout_rate=0.1,
            num_rnn_layers=1):
        """Initialize Cash Controller.

        :param int metafeature_size: dimensionality of metafeatures
        :param int input_size: dimensionality of input features (actions)
        :param int hidden_size: dimensionality of hidden RNN layer.
        :param int output_size: dimensionality of output space.
        :param AlgorithmSpace a_space:
        :param list[str] mlf_signature:
        :param int metafeature_encoding_size: dim of metafeatures embedding
        :param float dropout_rate:
        :param int num_rnn_layers:
        :ivar int aux_size: dimensionality of auxiliary inputs (reward and
            action from previous time-step).
        :ivar list[dict] action_classifiers:
        :ivar list[float] log_prob_buffer:
        :ivar list[float] reward buffer:
        """
        super(MetaLearnController, self).__init__()
        self.a_space = a_space
        self.dropout_rate = dropout_rate
        self.num_rnn_layers = num_rnn_layers
        self.metafeature_size = metafeature_size
        self.input_size = input_size

        # action dim for previous reward. Note that to implement meta-learning,
        # the previous action is encoded directly in the input.
        self.aux_reward_size = aux_reward_size
        self.hidden_size = hidden_size
        self.output_size = output_size
        self.metafeature_encoding_size = metafeature_encoding_size

        # an array storing classifiers for algorithm components and
        # hyperparameter settings
        self.action_classifiers = []
        # maps algorithm types to index pointer referring to action classifier
        # for algorithm components of that type.
        self.atype_map = {}
        # maps algorithm component to list of index pointers referring to
        # action classifier for hyperparameters for that component
        self.acomponent_to_hyperparams = defaultdict(list)
        # maps hyperparameter actions to conditional masks that determines what
        # subsequent actions the agent can make.
        self.hyperparam_exclude_conditions = defaultdict(dict)

        # buffers needed for backprop
        self.value_buffer = []
        self.log_prob_buffer = []
        self.reward_buffer = []
        self.entropy_buffer = []

        # architecture specification

        # meta-rnn models representation of previous action, previous reward,
        # and metafeatures
        self.meta_rnn = nn.GRU(
            self.input_size + self.aux_reward_size +
            self.metafeature_encoding_size,
            self.hidden_size,
            num_layers=self.num_rnn_layers,
            dropout=self.dropout_rate,
        )
        self.meta_decoder = nn.Linear(self.hidden_size, self.output_size)
        self.meta_dropout = nn.Dropout(self.dropout_rate)

        # special encoding layer for initial metafeature input
        self.metafeature_dense = nn.Linear(
            self.metafeature_size, self.metafeature_encoding_size)

        # critic layer that produces value estimate
        self.critic_dropout = nn.Dropout(self.dropout_rate)
        self.critic_linear1 = nn.Linear(self.output_size, self.hidden_size)
        self.critic_linear2 = nn.Linear(self.hidden_size, self.hidden_size)
        self.critic_linear3 = nn.Linear(self.hidden_size, self.hidden_size)
        self.critic_output = nn.Linear(self.hidden_size, 1)

        # micro action policy rnn takes meta-rnn representation and selects
        # hyperparameters to specify an ML Framework.
        self.micro_action_rnn = nn.GRU(
            self.input_size,
            self.hidden_size,
            num_layers=self.num_rnn_layers,
            dropout=self.dropout_rate,
        )
        self.micro_action_decoder = nn.Linear(
            self.hidden_size, self.output_size)
        self.micro_action_dropout = nn.Dropout(self.dropout_rate)

        # for each algorithm component and hyperparameter value, create a
        # softmax classifier over the number of unique components/hyperparam
        # values.
        self._mlf_signature = mlf_signature
        all_components = self.a_space.component_dict_from_signature(
            self.a_space.ALL_COMPONENTS if self._mlf_signature is None
            else self._mlf_signature)
        idx = 0

        for atype, acomponents in all_components.items():
            # action classifier
            self._add_action_classifier(
                idx, CASHComponent.ALGORITHM, atype, acomponents, None)
            self.atype_map[atype] = idx
            idx += 1
            for acomponent in acomponents:
                h_state_space = self.a_space.h_state_space([acomponent])
                h_exclude_cond_map = self.a_space.h_exclusion_conditions(
                    [acomponent])
                for hname, hchoices in h_state_space.items():
                    exclude_masks = self._exclusion_condition_to_mask(
                        h_state_space, h_exclude_cond_map.get(hname, None))
                    self._add_action_classifier(
                        idx, CASHComponent.HYPERPARAMETER, hname, hchoices,
                        exclude_masks)
                    self.acomponent_to_hyperparams[acomponent].append(idx)
                    idx += 1
        self.n_actions = len(self.action_classifiers)

    def _exclusion_condition_to_mask(self, h_state_space, exclude):
        """
        Create a mask of the action space to prevent selecting invalid states
        given previous choices. A mask value 1 means to exclude the value from
        the state space.
        """

        if exclude is None:
            return None

        exclude_masks = defaultdict(dict)
        for choice, exclude_conditions in exclude.items():
            for hname, exclude_values in exclude_conditions.items():
                if exclude_values == EXCLUDE_ALL:
                    mask = [1 for _ in h_state_space[hname]]
                else:
                    mask = [
                        int(i in exclude_values) for i in h_state_space[hname]]
                exclude_masks[choice].update({hname: torch.ByteTensor(mask)})

        return exclude_masks

    def _add_action_classifier(
            self, action_index, action_type, name, choices,
            exclude_masks):
        """Add action classifier to the nn.Module."""
        dense_attr = "action_dense_%s" % action_index
        softmax_attr = "action_softmax_%s" % action_index
        embedding_attr = "action_embedding_%s" % action_index
        # set layer attributes for action classifiers
        n_choices = len(choices)
        setattr(self, dense_attr, nn.Linear(self.output_size, n_choices))
        setattr(self, softmax_attr, nn.Softmax(1))
        setattr(self, embedding_attr, nn.Embedding(n_choices, self.input_size))
        # keep track of metadata for lookup purposes
        self.action_classifiers.append({
            "action_type": action_type,
            "name": name,
            "choices": choices,
            "dense_attr": dense_attr,
            "softmax_attr": softmax_attr,
            "embedding_attr": embedding_attr,
            "exclude_masks": exclude_masks,
        })

    def get_policy_dist(self, action, hidden, action_classifier, mask=None):
        """Get action distribution from policy."""
        rnn_output, hidden = self.micro_action_rnn(action, hidden)
        rnn_output = self.micro_action_dropout(
            self.micro_action_decoder(rnn_output))

        dense = getattr(self, action_classifier["dense_attr"])
        softmax = getattr(self, action_classifier["softmax_attr"])

        # obtain action probability distribution
        rnn_output = rnn_output.view(rnn_output.shape[0], -1)
        logits = dense(rnn_output)
        if mask is not None:
            logger.warning(
                "applying mask %s to action_classifier %s" %
                (mask, action_classifier))
            logits[mask] = float("-inf")
        action_probs = softmax(logits)
        return action_probs, hidden

    def get_value(self, state):
        """Get value estimate from state.

        The state includes dataset metafeatures, previous action, and previous
        reward.
        """
        value = self.critic_dropout(F.relu(self.critic_linear1(state)))
        value = self.critic_dropout(F.relu(self.critic_linear2(value)))
        value = self.critic_dropout(F.relu(self.critic_linear3(value)))
        return self.critic_output(value)

    def forward(
            self,
            prev_action: torch.FloatTensor,
            prev_reward: torch.FloatTensor,
            metafeatures: torch.FloatTensor,
            hidden: torch.FloatTensor,
            target_type: TargetType):
        """Decode a metafeature tensor to sequence of actions.

        Where the actions are a sequence of algorithm components and
        hyperparameter settings.

        :prev_action: tensor of action distribution at previous time step.
        :prev_reward: single-element tensor of reward at previous time step.
        :metafeatures: tensor of dataset feature representation.
        :hidden: hidden state of previous time step.
        :target_type: type of supervision target
        """
        actions = []

        # first process inputs with metalearning rnn
        state, hidden = self.meta_rnn(
            torch.cat(
                [
                    prev_action,
                    prev_reward,
                    F.relu(self.metafeature_dense(metafeatures))
                ],
                dim=2
            ),
            hidden)
        action_tensor = state = self.meta_dropout(
            self.meta_decoder(state))

        # get value estimate
        value = self.get_value(state)

        # create attribute dictionary that maps hyperparameter names to
        # conditional masks
        self._exclude_masks = {}
        # for each algorithm component type, select an algorithm
        if self._mlf_signature is None:
            algorithm_components = \
                self.a_space.component_dict_from_target_type(target_type)
        else:
            algorithm_components = \
                self.a_space.component_dict_from_signature(self._mlf_signature)

        for atype in algorithm_components:
            action, action_tensor, hidden = self._decode_action(
                action_tensor, hidden, action_index=self.atype_map[atype])
            if action is None:
                raise RuntimeError(
                    "selected action for algorithm component cannot be None.\n"
                    "exclude masks: %s\n"
                    "algorithm components: %s\n"
                    "algorithm type: %s" % (
                        self._exclude_masks,
                        algorithm_components,
                        atype,
                    )
                )
            actions.append(action)
            # each algorithm is associated with a set of hyperparameters
            for h_index in self.acomponent_to_hyperparams[action["choice"]]:
                hyperparameter_action, action_tensor, hidden = \
                    self._decode_action(
                        action_tensor, hidden, action_index=h_index)
                if hyperparameter_action is None:
                    logger.warning(
                        "selected action for hyperparameter index %s for "
                        "algorithm choice %s" % (h_index, action["choice"])
                    )
                    continue
                actions.append(hyperparameter_action)
        return value, actions, action_tensor, hidden

    def _get_exclusion_mask(self, action_name):
        if action_name in self._exclude_masks:
            return (
                self._exclude_masks[action_name]
                .view(1, -1).type(torch.bool)
            )
        else:
            return None

    def _accumulate_exclusion_mask(self, masks):
        for action_name, m in masks.items():
            mask = self._exclude_masks.get(action_name, None)
            if mask is None:
                self._exclude_masks[action_name] = m
            else:
                acc_mask = (m.int() + mask.int()) > 0
                acc_mask = acc_mask.type(torch.ByteTensor)
                self._exclude_masks[action_name] = acc_mask

    def _decode_action(
            self, action_tensor, hidden, action_index):
        action_classifier = self.action_classifiers[action_index]
        mask = self._get_exclusion_mask(action_classifier["name"])
        if mask is not None and (mask == 1).all():
            # If all actions are masked, then don't select any choice
            logger.warning(
                "all actions are masked for action_classifier %s" %
                action_classifier)
            return None, action_tensor, hidden
        action_probs, hidden = self.get_policy_dist(
            action_tensor, hidden, action_classifier, mask)
        action = self._select_action(action_probs, action_classifier, mask)
        if action is None:
            return None, action_tensor, hidden
        action_tensor = self.encode_embedding(
            action_index, action["choice_index"])
        return action, action_tensor, hidden

    def _select_action(self, action_probs, action_classifier, mask):
        """Select action based on action probability distribution."""
        # compute unmasked log probabilitises.
        policy_dist = Categorical(action_probs)
        choice_index = policy_dist.sample()
        _choice_index = choice_index.data.item()
        choice = action_classifier["choices"][_choice_index]

        # for current choice, accumulate exclusion masks if any
        if action_classifier["exclude_masks"] is not None and \
                choice in action_classifier["exclude_masks"]:
            for _, masks in action_classifier["exclude_masks"].items():
                self._accumulate_exclusion_mask(masks)

        return {
            "action_type": action_classifier["action_type"],
            "action_name": action_classifier["name"],
            "choices": action_classifier["choices"],
            "choice_index": _choice_index,
            "choice": choice,
            "log_prob": policy_dist.log_prob(choice_index),
            "entropy": policy_dist.entropy(),
        }

    def encode_embedding(self, action_index, choice_index):
        """Encode action choice into embedding at input for next action."""
        action_classifier = self.action_classifiers[action_index]
        embedding = getattr(self, action_classifier["embedding_attr"])
        action_embedding = embedding(torch.LongTensor([choice_index]))
        return action_embedding.view(
            1, action_embedding.shape[0], action_embedding.shape[1])

    def init_hidden(self):
        """Initialize hidden layer with zeros."""
        return torch.zeros(
            self.num_rnn_layers, 1, self.hidden_size, requires_grad=True)

    def init_action(self):
        """Initialize action input state."""
        return torch.zeros(1, 1, self.input_size, requires_grad=True)

    @property
    def config(self):
        return {
            "metafeature_size": self.metafeature_size,
            "input_size": self.input_size,
            "hidden_size": self.hidden_size,
            "output_size": self.output_size,
            "a_space": self.a_space,
            "dropout_rate": self.dropout_rate,
            "num_rnn_layers": self.num_rnn_layers,
        }

    def save(self, path):
        """Save weights and configuration."""
        torch.save(
            {
                "config": self.config,
                "weights": self.state_dict(),
            },
            path,
            pickle_module=dill)

    @classmethod
    def load(cls, path):
        """Load saved controller."""
        model_config = torch.load(path, pickle_module=dill)
        rnn = cls(**model_config["config"])
        rnn.load_state_dict(model_config["weights"])
        return rnn



"""Module for evaluating trained controller."""

import os

from pathlib import Path

import pandas as pd
import torch

from metalearn.metalearn_controller import MetaLearnController
from metalearn.inference.inference_engine import CASHInference
from metalearn.task_environment import TaskEnvironment


MODEL_PATH = Path(os.path.dirname(__file__)) / ".." / "floyd_outputs" / "290"
OUTPUT_PATH = Path.home() / ".metalearn-cache" / "inference_results" / "290"
OUTPUT_PATH.mkdir(parents=True, exist_ok=True)


def load_experiment_results():
    return pd.read_csv(
        MODEL_PATH / "rnn_metalearn_controller_experiment.csv")


def load_controller():
    return MetaLearnController.load(MODEL_PATH / "controller_trial_0.pt")


def load_eval_task_env():
    torch.manual_seed(10)
    return TaskEnvironment(
        env_sources=["SKLEARN", "OPEN_ML", "KAGGLE"],
        test_env_sources=["AUTOSKLEARN_BENCHMARK"],
        test_set_config={
            "AUTOSKLEARN_BENCHMARK": {
                "test_size": 0.8, "random_state": 100, "verbose": 5}
        },
        random_state=100,
        enforce_limits=True,
        per_framework_time_limit=720,
        per_framework_memory_limit=10000,
        dataset_names=None,
        test_dataset_names=None,
        error_reward=0,
        target_types=["BINARY"])


def _compute_inference_results(
        task_env, inference_engine, n_shots, n_eval_samples,
        evaluate_test_env=False):

    if evaluate_test_env:
        data_distribution = task_env.test_data_distribution
        inference_fn = inference_engine.evaluate_test_data_envs
    else:
        data_distribution = task_env.data_distribution
        inference_fn = inference_engine.evaluate_training_data_envs

    data_dist_inference_results = []
    for data_env in data_distribution:
        for i in range(n_eval_samples):
            inference_results = inference_fn(
                n=n_shots, datasets=[data_env.name], verbose=True)
            data_dist_inference_results.append(
                create_inference_result_df(inference_results)
                .assign(
                    data_env=data_env.name,
                    target_type=data_env.target_type.name,
                    n_eval_sample=i,
                )
            )
    data_dist_inference_results = \
        None if len(data_dist_inference_results) == 0 \
        else pd.concat(data_dist_inference_results)

    return data_dist_inference_results


def evaluate_controller(
        controller, task_env, meta_reward_multiplier, n_shots=20,
        n_eval_samples=5):

    inference_engine = CASHInference(
        controller, task_env, meta_reward_multiplier)

    # evaluate controller on training data envs
    train_env_inference_results = _compute_inference_results(
        task_env, inference_engine, n_shots, n_eval_samples,
    )

    # evaluate controller on test data environments
    test_env_inference_results = _compute_inference_results(
        task_env, inference_engine, n_shots, n_eval_samples,
        evaluate_test_env=True,
    )

    return {
        "training_env": train_env_inference_results,
        "test_env": test_env_inference_results,
    }


def create_inference_result_df(inference_results):

    def _tabulate_results(fn, key):
        df = pd.DataFrame({
            k: [fn(x) for x in v]
            for k, v in inference_results.items()
        })
        df = df.unstack().reset_index([0, 1])
        df.columns = ["data_env", "n_inference_steps", "value"]
        df["key"] = key
        return df[["data_env", "n_inference_steps", "key", "value"]]

    return pd.concat([
        _tabulate_results(lambda x: x.validation_score, "validation_score"),
        _tabulate_results(lambda x: x.reward, "reward"),
        _tabulate_results(lambda x: x.is_valid, "is_valid"),
        _tabulate_results(lambda x: x.mlf, "mlf"),
        _tabulate_results(lambda x: x.mlf_full, "mlf_full"),
        _tabulate_results(lambda x: x.scorer, "scorer"),
        _tabulate_results(lambda x: x.target_type, "target_type"),
    ])



from . import (
    algorithm_space,
    enforce_fit_limits,
    metalearn_controller,
    metalearn_reinforce,
    components,
    data_environments,
    data_types,
    errors,
    experiment,
    inference,
    loggers,
    plotting,
    random_search,
    scorers,
    task_environment,
    tracking)


__all__ = [
    "algorithm_space",
    "enforce_fit_limits",
    "metalearn_controller",
    "metalearn_reinforce",
    "components",
    "data_environments",
    "data_types",
    "errors",
    "inference",
    "experiment",
    "loggers",
    "plotting",
    "random_search",
    "scorers",
    "task_environment",
    "tracking",
]



"""Feature map for the kaggle competition 'costa rican household poverty'."""

from collections import OrderedDict

from ...data_types import FeatureType


def get_feature_map():
    return OrderedDict([
        ("v2a1", FeatureType.CONTINUOUS),
        ("hacdor", FeatureType.CATEGORICAL),
        ("rooms", FeatureType.CONTINUOUS),
        ("hacapo", FeatureType.CATEGORICAL),
        ("v14a", FeatureType.CATEGORICAL),
        ("refrig", FeatureType.CATEGORICAL),
        ("v18q", FeatureType.CATEGORICAL),
        ("v18q1", FeatureType.CONTINUOUS),
        ("r4h1", FeatureType.CONTINUOUS),
        ("r4h2", FeatureType.CONTINUOUS),
        ("r4h3", FeatureType.CONTINUOUS),
        ("r4m1", FeatureType.CONTINUOUS),
        ("r4m2", FeatureType.CONTINUOUS),
        ("r4m3", FeatureType.CONTINUOUS),
        ("r4t1", FeatureType.CONTINUOUS),
        ("r4t2", FeatureType.CONTINUOUS),
        ("r4t3", FeatureType.CONTINUOUS),
        ("tamhog", FeatureType.CONTINUOUS),
        ("tamviv", FeatureType.CONTINUOUS),
        ("escolari", FeatureType.CONTINUOUS),
        ("rez_esc", FeatureType.CONTINUOUS),
        ("hhsize", FeatureType.CONTINUOUS),
        ("paredblolad", FeatureType.CATEGORICAL),
        ("paredzocalo", FeatureType.CATEGORICAL),
        ("paredpreb", FeatureType.CATEGORICAL),
        ("pareddes", FeatureType.CATEGORICAL),
        ("paredmad", FeatureType.CATEGORICAL),
        ("paredzinc", FeatureType.CATEGORICAL),
        ("paredfibras", FeatureType.CATEGORICAL),
        ("paredother", FeatureType.CATEGORICAL),
        ("pisomoscer", FeatureType.CATEGORICAL),
        ("pisocemento", FeatureType.CATEGORICAL),
        ("pisoother", FeatureType.CATEGORICAL),
        ("pisonatur", FeatureType.CATEGORICAL),
        ("pisonotiene", FeatureType.CATEGORICAL),
        ("pisomadera", FeatureType.CATEGORICAL),
        ("techozinc", FeatureType.CATEGORICAL),
        ("techoentrepiso", FeatureType.CATEGORICAL),
        ("techocane", FeatureType.CATEGORICAL),
        ("techootro", FeatureType.CATEGORICAL),
        ("cielorazo", FeatureType.CATEGORICAL),
        ("abastaguadentro", FeatureType.CATEGORICAL),
        ("abastaguafuera", FeatureType.CATEGORICAL),
        ("abastaguano", FeatureType.CATEGORICAL),
        ("public", FeatureType.CATEGORICAL),
        ("planpri", FeatureType.CATEGORICAL),
        ("noelec", FeatureType.CATEGORICAL),
        ("coopele", FeatureType.CATEGORICAL),
        ("sanitario1", FeatureType.CATEGORICAL),
        ("sanitario2", FeatureType.CATEGORICAL),
        ("sanitario3", FeatureType.CATEGORICAL),
        ("sanitario5", FeatureType.CATEGORICAL),
        ("sanitario6", FeatureType.CATEGORICAL),
        ("energcocinar1", FeatureType.CATEGORICAL),
        ("energcocinar2", FeatureType.CATEGORICAL),
        ("energcocinar3", FeatureType.CATEGORICAL),
        ("energcocinar4", FeatureType.CATEGORICAL),
        ("elimbasu1", FeatureType.CATEGORICAL),
        ("elimbasu2", FeatureType.CATEGORICAL),
        ("elimbasu3", FeatureType.CATEGORICAL),
        ("elimbasu4", FeatureType.CATEGORICAL),
        ("elimbasu5", FeatureType.CATEGORICAL),
        ("elimbasu6", FeatureType.CATEGORICAL),
        ("epared1", FeatureType.CATEGORICAL),
        ("epared2", FeatureType.CATEGORICAL),
        ("epared3", FeatureType.CATEGORICAL),
        ("etecho1", FeatureType.CATEGORICAL),
        ("etecho2", FeatureType.CATEGORICAL),
        ("etecho3", FeatureType.CATEGORICAL),
        ("eviv1", FeatureType.CATEGORICAL),
        ("eviv2", FeatureType.CATEGORICAL),
        ("eviv3", FeatureType.CATEGORICAL),
        ("dis", FeatureType.CATEGORICAL),
        ("male", FeatureType.CATEGORICAL),
        ("female", FeatureType.CATEGORICAL),
        ("estadocivil1", FeatureType.CATEGORICAL),
        ("estadocivil2", FeatureType.CATEGORICAL),
        ("estadocivil3", FeatureType.CATEGORICAL),
        ("estadocivil4", FeatureType.CATEGORICAL),
        ("estadocivil5", FeatureType.CATEGORICAL),
        ("estadocivil6", FeatureType.CATEGORICAL),
        ("estadocivil7", FeatureType.CATEGORICAL),
        ("parentesco1", FeatureType.CATEGORICAL),
        ("parentesco2", FeatureType.CATEGORICAL),
        ("parentesco3", FeatureType.CATEGORICAL),
        ("parentesco4", FeatureType.CATEGORICAL),
        ("parentesco5", FeatureType.CATEGORICAL),
        ("parentesco6", FeatureType.CATEGORICAL),
        ("parentesco7", FeatureType.CATEGORICAL),
        ("parentesco8", FeatureType.CATEGORICAL),
        ("parentesco9", FeatureType.CATEGORICAL),
        ("parentesco10", FeatureType.CATEGORICAL),
        ("parentesco11", FeatureType.CATEGORICAL),
        ("parentesco12", FeatureType.CATEGORICAL),
        ("idhogar", FeatureType.CATEGORICAL),
        ("hogar_nin", FeatureType.CATEGORICAL),
        ("hogar_adul", FeatureType.CATEGORICAL),
        ("hogar_mayor", FeatureType.CATEGORICAL),
        ("hogar_total", FeatureType.CATEGORICAL),
        ("dependency", FeatureType.CONTINUOUS),
        ("edjefe", FeatureType.CONTINUOUS),
        ("edjefa", FeatureType.CONTINUOUS),
        ("meaneduc", FeatureType.CONTINUOUS),
        ("instlevel1", FeatureType.CATEGORICAL),
        ("instlevel2", FeatureType.CATEGORICAL),
        ("instlevel3", FeatureType.CATEGORICAL),
        ("instlevel4", FeatureType.CATEGORICAL),
        ("instlevel5", FeatureType.CATEGORICAL),
        ("instlevel6", FeatureType.CATEGORICAL),
        ("instlevel7", FeatureType.CATEGORICAL),
        ("instlevel8", FeatureType.CATEGORICAL),
        ("instlevel9", FeatureType.CATEGORICAL),
        ("bedrooms", FeatureType.CATEGORICAL),
        ("overcrowding", FeatureType.CATEGORICAL),
        ("tipovivi1", FeatureType.CATEGORICAL),
        ("tipovivi2", FeatureType.CATEGORICAL),
        ("tipovivi3", FeatureType.CATEGORICAL),
        ("tipovivi4", FeatureType.CATEGORICAL),
        ("tipovivi5", FeatureType.CATEGORICAL),
        ("computer", FeatureType.CATEGORICAL),
        ("television", FeatureType.CATEGORICAL),
        ("mobilephone", FeatureType.CATEGORICAL),
        ("qmobilephone", FeatureType.CONTINUOUS),
        ("lugar1", FeatureType.CATEGORICAL),
        ("lugar2", FeatureType.CATEGORICAL),
        ("lugar3", FeatureType.CATEGORICAL),
        ("lugar4", FeatureType.CATEGORICAL),
        ("lugar5", FeatureType.CATEGORICAL),
        ("lugar6", FeatureType.CATEGORICAL),
        ("area1", FeatureType.CATEGORICAL),
        ("area2", FeatureType.CATEGORICAL),
        ("age", FeatureType.CONTINUOUS),
        ("SQBescolari", FeatureType.CONTINUOUS),
        ("SQBage", FeatureType.CONTINUOUS),
        ("SQBhogar_total", FeatureType.CONTINUOUS),
        ("SQBedjefe", FeatureType.CONTINUOUS),
        ("SQBhogar_nin", FeatureType.CONTINUOUS),
        ("SQBovercrowding", FeatureType.CONTINUOUS),
        ("SQBdependency", FeatureType.CONTINUOUS),
        ("SQBmeaned", FeatureType.CONTINUOUS),
        ("agesq", FeatureType.CONTINUOUS),
    ])



"""Helper functions for plotly graphs."""

import colorlover
import math
import plotly.graph_objs as go

from collections import defaultdict
from plotly import tools

PALETTE = colorlover.scales["8"]["qual"]["Dark2"]
METRICS = [
    "losses",
    "aggregate_gradients",
    "best_validation_scores",
    "mean_rewards",
    "mean_validation_scores",
    "n_successful_mlfs",
    "mlf_diversity",
    "hyperparam_diversity",
]


def subplot_coords(iterable, ncols, one_indexed=True):
    n = len(iterable)
    nrows = math.ceil(n / ncols)
    offset = 1 if one_indexed else 0
    return {
        "nrows": nrows,
        "ncols": ncols,
        "coords":  [(i + offset, j + offset)
                    for i in range(nrows)
                    for j in range(ncols)]
    }


def create_time_series(x, y, group_name, showlegend, group_colormap=None):
    line_dict = dict(width=1)
    if group_colormap is not None:
        line_dict.update(dict(color=group_colormap[group_name]))
    return go.Scatter(
        x=x,
        y=y,
        name=group_name,
        legendgroup=group_name,
        mode='lines',
        line=line_dict,
        showlegend=showlegend,
        opacity=0.7,
    )


def create_multi_time_series(
        results, group_column, metric, legend_metric="mlf_diversity"):
    groups = results[group_column].unique()
    cm = {g: PALETTE[i] for i, g in enumerate(groups)}
    showlegend = True if metric == legend_metric else False
    return (
        results
        .groupby(group_column)
        .apply(lambda df: create_time_series(
            df["episode"],
            df[metric],
            df[group_column].iloc[0],
            showlegend,
            cm
        ))
        .tolist())


def plot_run_history(results):
    coords = subplot_coords(METRICS, 2)
    fig = tools.make_subplots(
        rows=coords["nrows"],
        cols=coords["ncols"],
        subplot_titles=METRICS,
        vertical_spacing=0.1,
        print_grid=False)

    for i, metric in enumerate(METRICS):
        traces = create_multi_time_series(
            results.astype({"job_number": str}),
            "job_number",
            metric,
        )
        row_i = coords["coords"][i][0]
        col_i = coords["coords"][i][1]
        for trace in traces:
            fig.append_trace(trace, row_i, col_i)
        # add x-axis titles on the bottom ncols plots
        if i >= (coords["ncols"] * coords["nrows"] - coords["ncols"]):
            xax = "xaxis%s" % ("" if i == 0 else i + 1)
            fig.layout[xax].update({"title": "episode"})

    fig.layout.update({
        "height": 600,
    })
    for annotation in fig.layout.annotations:
        annotation.font.update({"size": 12})
    return fig


def plot_run_history_by_dataenv(results, metric="mean_rewards", ncols=3):

    colormap = {
        g: PALETTE[i] for i, g in
        enumerate(results.job_number.unique())}

    def time_series(df, y, legend_metric="anneal"):
        line_dict = dict(width=1)
        job_number = df["job_number"].iloc[0]
        env_name = df["data_env_names"].iloc[0]
        color = colormap.get(job_number)
        showlegend = True if env_name == legend_metric else False
        if color is not None:
            line_dict.update(dict(color=color))
        return go.Scatter(
            x=df["episode"],
            y=df[y],
            name=str(job_number),
            legendgroup=str(job_number),
            line=line_dict,
            mode='lines',
            showlegend=showlegend,
        )

    # time_series_data is a dict where the key is
    # the env_name and value is the corresponding
    # trace.
    _time_series_data = (
        results
        .groupby(["data_env_names", "job_number"])
        .apply(time_series, y=metric)
        .to_dict()
    )

    time_series_data = defaultdict(dict)
    for (data_env, job_num), trace in _time_series_data.items():
        time_series_data[data_env][job_num] = trace

    coords = subplot_coords(time_series_data, ncols)
    fig = tools.make_subplots(
        rows=coords["nrows"],
        cols=coords["ncols"],
        subplot_titles=list(time_series_data.keys()),
        vertical_spacing=0.05,
        print_grid=False)

    for i, (data_env, traces) in enumerate(time_series_data.items()):
        row_i, col_i = coords["coords"][i][0], coords["coords"][i][1]
        for job_num, trace in traces.items():
            fig.append_trace(trace, row_i, col_i)
            # add x-axis titles on the bottom ncols plots
            if i >= (coords["ncols"] * coords["nrows"] - coords["ncols"]):
                xax = "xaxis%s" % ("" if i == 0 else i + 1)
                fig.layout[xax].update({"title": "episode"})

    fig.layout.update({
        "height": 1000,
    })

    for annotation in fig.layout.annotations:
        annotation.font.update({"size": 12})

    return fig


def plot_best_mlfs(best_mlfs):
    def create_best_mlf_timeline(x, y, color):
        return go.Scatter(
            x=x,
            y=y,
            mode='markers',
            opacity=0.7,
            line=dict(width=1, color=color)
        )

    colormap = {
        g: PALETTE[i] for i, g in
        enumerate(best_mlfs.job_number.unique())}

    traces = (
        best_mlfs.groupby("job_number")
        .apply(lambda df: create_best_mlf_timeline(
            df.episode, df.mlf, colormap.get(df.name)))
    ).tolist()

    fig = go.Figure(
        data=traces,
        layout=dict(
            height=500,
            margin=dict(l=600),
            hovermode="closest"
        ))

    fig.layout["xaxis"].update({"title": "episode"})
    return fig



"""Source datasets from OpenML.

https://www.openml.org/search?type=data
"""

import logging
import os

from collections import OrderedDict
from functools import partial
from pathlib import Path

import numpy as np
import openml

from openml import config
from openml.exceptions import OpenMLServerException, OpenMLPrivateDatasetError
from typing import Dict, FrozenSet

from .data_environment import DataEnvironment
from . import autosklearn_clf_task_ids
from ..data_types import FeatureType, TargetType, OpenMLTaskType, \
    DataSourceType


config.set_cache_directory(
    os.getenv("OPEN_ML_CACHE_DIR", Path.home() / ".openml" / "cache"))

logger = logging.getLogger(__name__)


N_CLASSIFICATION_ENVS = 20
N_REGRESSION_ENVS = 20
SPARSE_DATA_FORMAT = "Sparse_ARFF"


# Map OpenML feature data types to metalearn feature data types
FEATURE_TYPE_MAP = {
    "nominal": FeatureType.CATEGORICAL,
    "numeric": FeatureType.CONTINUOUS,
}
DATASOURCE_TYPES = [
    DataSourceType.OPEN_ML,
    DataSourceType.AUTOSKLEARN_BENCHMARK,
    DataSourceType.OPEN_ML_BENCHMARK_CC18,
]

AUTOSKLEARN_BENCHMARK_TASK_IDS = frozenset(autosklearn_clf_task_ids.TASK_IDS)


def _map_feature(feature):
    feature_dtype = FEATURE_TYPE_MAP.get(feature.data_type, None)
    if feature_dtype is None:
        raise ValueError(
            "feature dtype %s not recognized. valid values are %s" % (
                FEATURE_TYPE_MAP.keys()))
    return feature_dtype


def _normalize_feature_indices(feature_indices):
    min_index = min(feature_indices)
    return [i - min_index for i in feature_indices]


def openml_to_data_env(
        openml_dataset, target_column, target_type,
        data_source_type, test_size, random_state, target_preprocessor=None,
        include_features_with_na=False):
    print("fetching openml dataset id=%d: '%s'" %
          (openml_dataset.dataset_id, openml_dataset.name))
    if openml_dataset.format == SPARSE_DATA_FORMAT:
        raise TypeError("cannot handle sparse data.")
    if data_source_type not in DATASOURCE_TYPES:
        raise ValueError(
            "expected one of %s, found : %s" % (
                DATASOURCE_TYPES, data_source_type))
    feature_indices = []
    feature_types = []
    target_index = None
    ignore_atts = openml_dataset.ignore_attributes
    row_id_att = openml_dataset.row_id_attribute
    for key, feature in openml_dataset.features.items():
        if (ignore_atts and feature.name in ignore_atts) or \
                (row_id_att and feature.name == row_id_att):
            print("ignoring attribute %s in dataset %s" %
                  (ignore_atts, openml_dataset.name))
            continue

        if feature.name == target_column:
            target_index = key
            continue
        if feature.number_missing_values != 0:
            if include_features_with_na:
                feature_indices.append(key)
                feature_types.append(_map_feature(feature))
            else:
                continue
        else:
            feature_indices.append(key)
            feature_types.append(_map_feature(feature))

    if target_index is None:
        print("target column %s not found for dataset %s, skipping." %
              (target_column, openml_dataset.name))
        return None
    if len(feature_indices) == 0:
        print("no features found for dataset %s, skipping." %
              openml_dataset.name)
        return None

    def _fetch_training_data():
        # see https://openml.github.io/openml-python/develop/generated/openml.OpenMLDataset.html#openml.OpenMLDataset  # noqa E53
        # for more details
        data, *_ = openml_dataset.get_data(
            include_row_id=True, include_ignore_attributes=True,
            dataset_format="array")

        # remove rows with null targets. Only case we'll need that is if
        # metalearn supports semi-supervised learning.
        X = data[:, feature_indices]
        y = data[:, target_index].ravel()
        y_is_nan = np.isnan(y)
        n_nan = y_is_nan.sum()
        if n_nan > 0:
            print(
                "removing %d/%d rows with null targets" % (n_nan, X.shape[0])
            )
        return X[~y_is_nan, :], y[~y_is_nan]

    # normalize feature indices so that index is aligned with
    # _fetch_training_data
    min_index = min(feature_indices)
    normalized_feature_indices = [i - min_index for i in feature_indices]

    return DataEnvironment(
        name="openml.%s" % openml_dataset.name.lower(),
        source=data_source_type,
        target_type=target_type,
        feature_types=feature_types,
        feature_indices=normalized_feature_indices,
        # include row id ensures that indices are correctly aligned
        fetch_training_data=_fetch_training_data,
        fetch_test_data=None,
        test_size=test_size,
        random_state=random_state,
        target_preprocessor=target_preprocessor,
        scorer=None,
    )


def _get_target_type(num_classes, task_type):
    if task_type is OpenMLTaskType.SUPERVISED_REGRESSION:
        return TargetType.REGRESSION
    elif task_type is OpenMLTaskType.SUPERVISED_CLASSIFICATION:
        if num_classes > 2:
            return TargetType.MULTICLASS
        else:
            return TargetType.BINARY
    else:
        raise ValueError(
            "task_type %s is currently not recognized by metalearn." %
            task_type)


def _get_dataset_metadata(openml_task_metadata, task_type):
    target_columns, target_types = zip(*[
        (v["target_feature"],
         _get_target_type(v.get("NumberOfClasses", 0), task_type))
        for v in openml_task_metadata.values()])
    dataset_ids = [v["source_data"] for v in openml_task_metadata.values()]
    datasets, target_columns_out, target_types_out = [], [], []
    for id, tcols, ttypes in zip(dataset_ids, target_columns, target_types):
        try:
            datasets.append(openml.datasets.get_dataset(id))
            target_columns_out.append(tcols)
            target_types_out.append(ttypes)
        except OpenMLServerException as e:
            print("error when fetching openml metadata: %s, skipping dataset"
                  % e)
    return datasets, target_columns_out, target_types_out


def _create_envs(
        data_source_type, datasets, target_columns, target_types, test_size,
        random_state):
    _openml_to_data_env = partial(
        openml_to_data_env,
        data_source_type=data_source_type,
        test_size=test_size,
        random_state=random_state,
        include_features_with_na=True)
    envs = []
    for dataset, target_cols, target_types in zip(
            datasets, target_columns, target_types):
        try:
            envs.append(
                _openml_to_data_env(dataset, target_cols, target_types))
        except TypeError as e:
            print("skipping dataset %s due to error parsing dataset: %s" % (
                dataset.name, e))
    return OrderedDict([(d.name, d) for d in envs if d])


def _filter_out_ids(openml_task_metadata: Dict, filter_out_ids: FrozenSet):
    return {
        k: v for k, v in openml_task_metadata.items()
        if k not in filter_out_ids
    }


def classification_envs(
        n=N_CLASSIFICATION_ENVS, test_size=None, random_state=None,
        verbose=False):
    print("getting OpenML classification datasets")
    if n is None:
        n = N_CLASSIFICATION_ENVS
    task_type = OpenMLTaskType.SUPERVISED_CLASSIFICATION

    dataset_metadata = _get_dataset_metadata(
        _filter_out_ids(
            openml.tasks.list_tasks(task_type_id=task_type.value, size=n),
            # exclude any benchmark tasks
            AUTOSKLEARN_BENCHMARK_TASK_IDS.union(
                set(openml.study.get_suite('OpenML-CC18').tasks)
            )),
        task_type)
    return _create_envs(
        DataSourceType.OPEN_ML, *dataset_metadata +
        (test_size, random_state))


def regression_envs(
        n=N_REGRESSION_ENVS, test_size=None, random_state=None, verbose=False):
    print("getting OpenML regression datasets")
    if n is None:
        n = N_CLASSIFICATION_ENVS
    task_type = OpenMLTaskType.SUPERVISED_REGRESSION
    dataset_metadata = _get_dataset_metadata(
        _filter_out_ids(
            openml.tasks.list_tasks(task_type_id=task_type.value, size=n),
            # exclude any benchmark tasks
            AUTOSKLEARN_BENCHMARK_TASK_IDS.union(
                set(openml.study.get_suite('OpenML-CC18').tasks)
            )),
        task_type)
    return _create_envs(
        DataSourceType.OPEN_ML, *dataset_metadata + (test_size, random_state))


def _task_ids(n):
    if n is None:
        return autosklearn_clf_task_ids.TASK_IDS
    return autosklearn_clf_task_ids.TASK_IDS[:n]


def _get_task(id, verbose):
    if verbose:
        print("getting task metadata for task id %d" % id)
    try:
        task = openml.tasks.get_task(id)
        if task is None:
            print("skipping task id %d. "
                  "openml.tasks.get_task return null" % id)
        return task
    except (OpenMLServerException,
            OpenMLPrivateDatasetError,
            ValueError,
            TypeError) as e:
        print("skipping task id %d. error: %s" % (id, e))
        return None


def autosklearn_paper_classification_envs(
        n=None, test_size=None, random_state=None, verbose=False):
    """From Feurer et al. "Efficient and Robust Automated Machine Learning.

    https://papers.nips.cc/paper/5872-efficient-and-robust-automated-machine-learning  # noqa
    """
    print("getting OpenML Auto-SKlearn benchmark datasets")
    if n is None:
        task_ids = autosklearn_clf_task_ids.TASK_IDS
    else:
        task_ids = autosklearn_clf_task_ids.TASK_IDS[:n]
    # don't include other benchmark ids
    exclude_ids = set(openml.study.get_suite('OpenML-CC18').tasks)
    openml_tasks = [
        task_id for task_id in
        (_get_task(id, verbose) for id in task_ids)
        if task_id is not None and task_id not in exclude_ids
    ]
    openml_datasets = [t.get_dataset() for t in openml_tasks]
    target_columns = (t.target_name for t in openml_tasks)
    target_types = (
        _get_target_type(
            len(t.class_labels),
            OpenMLTaskType.SUPERVISED_CLASSIFICATION)
        for t in openml_tasks
        if t.class_labels is not None
    )
    return _create_envs(
        DataSourceType.AUTOSKLEARN_BENCHMARK, openml_datasets, target_columns,
        target_types, test_size, random_state)


def openml_benchmark_cc18_envs(
        n=None, test_size=None, random_state=None, verbose=False):
    """From OpenML benchmarks documentation:

    https://docs.openml.org/benchmark
    """
    print("getting OpenML-CC18 benchmark datasets")
    benchmark_suite = openml.study.get_suite('OpenML-CC18')
    if n is None:
        task_ids = list(benchmark_suite.tasks)
    else:
        task_ids = list(benchmark_suite.tasks)[:n]
    openml_tasks = list(filter(
        None, (_get_task(id, verbose) for id in task_ids)))
    openml_datasets = [t.get_dataset() for t in openml_tasks]
    target_columns = (t.target_name for t in openml_tasks)
    target_types = (
        _get_target_type(
            len(t.class_labels),
            OpenMLTaskType.SUPERVISED_CLASSIFICATION)
        for t in openml_tasks
        if t.class_labels is not None
    )
    return _create_envs(
        DataSourceType.OPEN_ML_BENCHMARK_CC18, openml_datasets, target_columns,
        target_types, test_size, random_state)



from setuptools import setup

with open('README.md') as f:
    long_description = f.read()


setup(
    name="meta-ml",
    version="0.0.21",
    description="MetaRL-based Estimator using Task-encodings for AutoML",
    long_description=long_description,
    long_description_content_type='text/markdown',
    url="https://github.com/cosmicBboy/ml-research/tree/master/metalearn",
    packages=[
        "metalearn",
        "metalearn.components",
        "metalearn.data_environments",
        "metalearn.data_environments.feature_maps",
        "metalearn.inference"],
    install_requires=[
        "colorlover",
        "click==7.0",
        "dill==0.3.1.1",
        "kaggle",
        "matplotlib",
        "numpy",
        "openml==0.9.0",
        "pandas==1.0.5",
        "plotly",
        "psutil",
        "torch==1.5.0",
        "scipy",
        "yamlordereddictloader",
    ],
    extras_require={
        "sklearn": [
            "scikit-learn>=0.21.2",
        ]
    },
    scripts=["bin/metalearn"],
)



"""Kaggle Base Class."""

import os
import pandas as pd
import subprocess

from pathlib import Path

from .data_environment import DataEnvironment
from ..data_types import DataSourceType, FeatureType


KAGGLE_CACHE_DIR = os.environ.get("KAGGLE_CACHE_DIR", "~/.kaggle/cache")
KAGGLE_COMPETITION_URL = "https://www.kaggle.com/c"


class KaggleCompetition(object):
    """API for accessing kaggle competition data."""

    def __init__(
            self,
            competition_id,
            features,
            target,
            training_data_fname,
            test_data_fname,
            scorer=None,
            file_format="csv",
            cache=KAGGLE_CACHE_DIR,
            custom_preprocessor=None):
        """Initialize Kaggle Competition object.

        :params str competition_id: kaggle identifier for the competition. This
            is a dash-delimited string, e.g. "allstate-claims-severity"
        :params dict[str -> FeatureType] features: a dictionary mapping feature
            column names to FeatureTypes.
        :params dict[str -> TargetType] target: a dictionary mapping target
            column to TargetType.
        :params str training_data_fname: filename of the training set.
        :params str test_data_fname: filename of the test set.
        :params Scorer|None scorer: a `scorer.Scorer` named tuple that
            specifies how to evaluate the performance of the dataset.
        :params str file_format: currently only csv files are supported.
        :params str cache: directory location for caching competition files.
        :param callable|None custom_preprocessor: a function for doing custom
            pre-processing on the training set. This function should take
            as an argument a dataframe (containing the training/test data) and
            return a dataframe with additional columns. This function should
            be specified for competition datasets with more complex structures
            that involve joining the training/test sets with an auxiliary data
            source.
        """
        self._competition_id = competition_id
        self._features = features
        self._target = target
        self._training_data_fname = training_data_fname
        self._test_data_fname = test_data_fname
        self._scorer = scorer
        if file_format != "csv":
            raise ValueError(
                "'%s' is not a valid file format, only '.csv' currently "
                "supported.")
        self._file_format = file_format
        self._cache = Path(cache).expanduser()
        self._custom_preprocessor = custom_preprocessor

        if not self._cache.exists():
            self._cache.mkdir()

        self._feature_names, self._feature_types = map(
            list, zip(*self._features.items()))

        # currently only support one target.
        if len(self._target) > 1:
            raise ValueError(
                "can only specify one target, found %s" % self.target.keys())
        self._target_name = list(self._target.keys())[0]
        self._target_type = self._target[self._target_name]

    def _download_and_cache(self):
        """Download cache to local drive."""
        if not self._dataset_filepath.exists():
            self._dataset_filepath.mkdir()
        if not (self._training_set_filepath.exists() and
                self._test_set_filepath.exists()):
            subprocess.call(
                self._download_api_cmd(), cwd=self._dataset_filepath)

    @property
    def dataset_name(self):
        return "kaggle.%s" % self._competition_id.replace("-", "_")

    @property
    def url(self):
        """Return url to kaggle dataset."""
        return "%s/%s" % (KAGGLE_COMPETITION_URL, self._competition_id)

    @property
    def scorer(self):
        """Return Scorer object."""
        return self._scorer

    @property
    def _dataset_filepath(self):
        return self._cache / self._competition_id

    @property
    def _training_set_filepath(self):
        return self._dataset_filepath / self._training_data_fname

    @property
    def _test_set_filepath(self):
        return self._dataset_filepath / self._test_data_fname

    @property
    def _datetime_features(self):
        return [k for k, v in self._features.items() if v is FeatureType.DATE]

    def _download_api_cmd(self):
        """Download and cache the competition files in cache directory.

        Note: this command line call checks whether dataset already downloaded
        and skips if true.
        """
        return [
            "kaggle",
            "competitions",
            "download",
            self._competition_id,
        ]

    def get_training_data(self, n_samples=None):
        """Return a dataframe containing the training set.

        :returns: tuple of arrays of training data. First element is an array
            of features, second is an array of targets.
        """
        dataset = pd.read_csv(
            self._training_set_filepath, parse_dates=self._datetime_features,
            nrows=n_samples)

        if self._custom_preprocessor:
            dataset = self._custom_preprocessor(dataset)

        feature_data = dataset[self._feature_names].values
        target_data = dataset[[i for i in self._target]].values

        if target_data.shape[1] == 1:
            target_data = target_data.ravel()
        return feature_data, target_data

    def get_test_data(self, n_samples=None):
        """Return a dataframe containing test set.

        Note that this dataset does not contain targets because it is used for
        evaluation on the kaggle platform. This method should only be used
        when evaluating the final test set performance of a model after
        hyperparameter tuning.

        :returns: pandas.DataFrame of test data. Only includes features.
        """
        dataset = pd.read_csv(
            self._test_set_filepath, parse_dates=self._datetime_features,
            nrows=n_samples)

        if self._custom_preprocessor:
            dataset = self._custom_preprocessor(dataset)

        feature_data = dataset[self._feature_names].values
        return feature_data, None

    def data_env(self, n_samples=None, test_size=None, random_state=None):
        """Convert kaggle competition to data environment."""
        self._download_and_cache()
        # TODO: if called with test_size arg, override the fetch_test_data
        # None and specify test_size and random_state. This should be used
        # for partitioning a holdout test set from the training set. Still
        # need to support `make_submission` and `get_submission_performance`
        # methods below (should probably be pull those out into kaggle_api as
        # their own functions).
        if test_size:
            fetch_test_data = None
        else:
            fetch_test_data = self.get_test_data
        return DataEnvironment(
            name=self.dataset_name,
            source=DataSourceType.KAGGLE,
            target_type=self._target_type,
            feature_types=self._feature_types,
            feature_indices=[i for i in range(len(self._feature_types))],
            fetch_training_data=self.get_training_data,
            fetch_test_data=fetch_test_data,
            test_size=test_size,
            random_state=random_state,
            target_preprocessor=None,
            scorer=self.scorer,
        )

    def make_submission(self):
        """Submit predictions to kaggle platform for evaluation."""
        # TODO: this requires a custom function with the signature:
        # (predictions) -> submission object
        # where submission object is in json or csv format and is
        # written to a temporary file, uploaded via
        # `kaggle competitions submit` cli command and then deleted.
        # This should also generate a unique submission
        # id that we can use when getting submission performance
        pass

    def get_submission_performance(self, submission_id):
        # TODO: this should take the output of `make_submission` and use the
        # `kaggle competitions submissions` command to get performance metric
        # for a particular submission.
        pass
